import os
import unittest

import bitcoin
from bitcoin import sha256, compress, privtopub, mk_multisig_script


class TestCustomTestSignet(unittest.TestCase):
    def setUp(self) -> None:
        self.test_signet_params = [31, 63, 20]  # addr, p2sh, priv

    def test_custom_signet_bip32(self):
        seed = b'custom signet bip32'
        print('doing')
        priv = bitcoin.bip32_master_key(seed, vbytes=bitcoin.deterministic.CUSTOM_SIGNET_TEST_PRIVATE)
        for path_element in [44 + 2 ** 31, 2 + 2 ** 31]:
            priv = bitcoin.bip32_ckd(priv, path_element)
        self.assertEqual(
            priv,
            'tprsZ9ywKur445iQcrHVg5LYS2jZHvK26RLvGPUiDtTuDmYV5vXGk2'
            'BpXCPzKRwvLJyTKm1diPYWhDRRALqfsE3uoNcA7Z7GJNJ96Nai4q3fAKk'
        )
        pub = bitcoin.bip32_privtopub(priv)
        self.assertEqual(
            pub,
            'tpqcXWEW4DmTkxsqr3pJWNdETrsxYjUYLfSi3GxWrmsQ8c1oQs1z4i'
            'DBrqXgUbMTyugnTC3vuAnxHNVxFH7BamKwVSHEzXMchi7k6oa2p6bHWyc'
        )

    def test_b58_p2pkh(self):
        priv1 = sha256(b'priv_key_text')
        pub1 = compress(privtopub(priv1))
        address = bitcoin.pubtoaddr(pub1, magicbyte=self.test_signet_params[0])
        self.assertEqual(address, 'DViH9eehQSZ41NQA9CLhpCPP1WWxJpNshH')
        scriptPubKey = bitcoin.address_to_script(address)
        self.assertEqual(
            address,
            bitcoin.script_to_address(scriptPubKey, self.test_signet_params[0])
        )
        self.assertEqual(scriptPubKey, '76a9140d899a3d4504d1cd493f822ae9ec6bdcabe092bb88ac')

    def test_b58_p2sh(self):
        priv1 = sha256(b'priv_key_text')
        priv2 = sha256(b'priv_key_text_2')
        priv3 = sha256(b'priv_key_text_3')
        pub1 = compress(privtopub(priv1))
        pub2 = compress(privtopub(priv2))
        pub3 = compress(privtopub(priv3))
        witness_program = mk_multisig_script([pub1, pub2, pub3], 2, 3)
        address = bitcoin.p2sh_scriptaddr(witness_program, self.test_signet_params[1])
        self.assertEqual(address, 'SMJgdd7bDnDKByBZdmrgoCu8TAUNypAqjE')
        script_p2sh = bitcoin.address_to_script(address)
        deserialized = bitcoin.deserialize_script(script_p2sh)
        self.assertEqual(deserialized[1], bitcoin.hash160(bytes.fromhex(witness_program)))
        self.assertEqual(script_p2sh, 'a91400210e5fea6ad1574d8cab58478a9f7c43f23bf587')

    def test_priv_wif(self):
        priv = sha256(b'priv_key_text')
        wif = bitcoin.encode_privkey(priv, 'wif', vbyte=self.test_signet_params[2])
        self.assertEqual(wif, '5yEK9ezuAC9tRBhokKKtXeLHzCEL5tNALdYQPsaiQBJwMgtWBCD')
        self.assertEqual(bitcoin.encode_privkey(wif, 'hex'), priv)

    def test_bf_p2sh(self):
        for _ in range(0, 10000):
            if not _ % 1000:
                print(_)
            priv1 = sha256(os.urandom(32))
            priv2 = sha256(b'priv_key_text_2')
            priv3 = sha256(b'priv_key_text_3')
            pub1 = compress(privtopub(priv1))
            pub2 = compress(privtopub(priv2))
            pub3 = compress(privtopub(priv3))
            witness_program = mk_multisig_script([pub1, pub2, pub3], 2, 3)
            address = bitcoin.p2sh_scriptaddr(witness_program, self.test_signet_params[1])
            self.assertTrue(address.startswith('S'), msg=address)
            script_p2sh = bitcoin.address_to_script(address)
            deserialized = bitcoin.deserialize_script(script_p2sh)
            self.assertEqual(deserialized[1], bitcoin.hash160(bytes.fromhex(witness_program)))

    def test_bf_p2pkh(self):
        for _ in range(0, 10000):
            if not _ % 1000:
                print(_)
            priv1 = sha256(os.urandom(32))
            pub1 = compress(privtopub(priv1))
            address = bitcoin.pubtoaddr(pub1, magicbyte=self.test_signet_params[0])
            self.assertTrue(address.startswith('D'), msg=address)
            scriptPubKey = bitcoin.address_to_script(address)
            self.assertEqual(
                address,
                bitcoin.script_to_address(scriptPubKey, self.test_signet_params[0])
            )

    def test_bf_wif_priv(self):
        for _ in range(0, 1000):
            if not _ % 1000:
                print(_)
            priv = sha256(os.urandom(32))
            wif = bitcoin.encode_privkey(priv, 'wif', vbyte=self.test_signet_params[2])
            self.assertTrue(wif.startswith('5'), msg=wif)
            self.assertEqual(bitcoin.encode_privkey(wif, 'hex'), priv)


class TestCustomSignet(unittest.TestCase):
    def setUp(self) -> None:
        self.test_signet_params = [115, 123, 82]  # addr, p2sh, priv

    def test_custom_signet_bip32(self):
        seed = b'custom signet bip32'
        print('doing')
        priv = bitcoin.bip32_master_key(seed, vbytes=bitcoin.deterministic.CUSTOM_SIGNET_PRIVATE)
        for path_element in [44 + 2 ** 31, 2 + 2 ** 31]:
            priv = bitcoin.bip32_ckd(priv, path_element)
        self.assertEqual(
            priv,
            'xprg83CBNA3hpdiSFUk45g268i3CcEVHMHzVGPX8YwSTrwnxVFpR'
            '4NNzPs3kPQVtmWBTD5JMmp3xYDBzKMcbAHJQwiaetK8w2JqNaXGEMZ1c3ck'
        )
        pub = bitcoin.bip32_privtopub(priv)
        self.assertEqual(
            pub,
            'xpuqhJUZiindjShTtMbJWHiUL2frbALb2n1QHjYtMeMXcQUZPj3z'
            'LcrwUdKDGoRmLv3LEnPqCZKMzWQB9f5GJbg23Cp6F7BucakeCypR7uakqJG'
        )

    def test_b58_p2pkh(self):
        priv1 = sha256(b'priv_key_text')
        pub1 = compress(privtopub(priv1))
        address = bitcoin.pubtoaddr(pub1, magicbyte=self.test_signet_params[0])
        self.assertEqual(address, 'oJ6wrmgu3bUehn7SCSKUYjFTstAC2TsKi4')
        scriptPubKey = bitcoin.address_to_script(address)
        self.assertEqual(
            address,
            bitcoin.script_to_address(scriptPubKey, self.test_signet_params[0])
        )
        self.assertEqual(scriptPubKey, '76a9140d899a3d4504d1cd493f822ae9ec6bdcabe092bb88ac')

    def test_b58_p2sh(self):
        priv1 = sha256(b'priv_key_text')
        priv2 = sha256(b'priv_key_text_2')
        priv3 = sha256(b'priv_key_text_3')
        pub1 = compress(privtopub(priv1))
        pub2 = compress(privtopub(priv2))
        pub3 = compress(privtopub(priv3))
        witness_program = mk_multisig_script([pub1, pub2, pub3], 2, 3)
        address = bitcoin.p2sh_scriptaddr(witness_program, self.test_signet_params[1])
        self.assertEqual(address, 'rVasi8zspc1tFyYm6wqotjELDRvyeMQt2i')
        script_p2sh = bitcoin.address_to_script(address)
        deserialized = bitcoin.deserialize_script(script_p2sh)
        self.assertEqual(deserialized[1], bitcoin.hash160(bytes.fromhex(witness_program)))
        self.assertEqual(script_p2sh, 'a91400210e5fea6ad1574d8cab58478a9f7c43f23bf587')

    def test_priv_wif(self):
        priv = sha256(b'priv_key_text')
        wif = bitcoin.encode_privkey(priv, 'wif', vbyte=self.test_signet_params[2])
        self.assertEqual(wif, '83kU62psza5vn3jXxFoqfXWJmGQrPYcT8Fw5XAEJ2oiyxH3jMEm')
        self.assertEqual(bitcoin.encode_privkey(wif, 'hex'), priv)

    def test_bf_p2sh(self):
        for _ in range(0, 10000):
            if not _ % 1000:
                print(_)
            priv1 = sha256(os.urandom(32))
            priv2 = sha256(b'priv_key_text_2')
            priv3 = sha256(b'priv_key_text_3')
            pub1 = compress(privtopub(priv1))
            pub2 = compress(privtopub(priv2))
            pub3 = compress(privtopub(priv3))
            witness_program = mk_multisig_script([pub1, pub2, pub3], 2, 3)
            address = bitcoin.p2sh_scriptaddr(witness_program, self.test_signet_params[1])
            self.assertTrue(address.startswith('r'), msg=address)
            script_p2sh = bitcoin.address_to_script(address)
            deserialized = bitcoin.deserialize_script(script_p2sh)
            self.assertEqual(deserialized[1], bitcoin.hash160(bytes.fromhex(witness_program)))

    def test_bf_p2pkh(self):
        for _ in range(0, 10000):
            if not _ % 1000:
                print(_)
            priv1 = sha256(os.urandom(32))
            pub1 = compress(privtopub(priv1))
            address = bitcoin.pubtoaddr(pub1, magicbyte=self.test_signet_params[0])
            self.assertTrue(address.startswith('o'), msg=address)
            scriptPubKey = bitcoin.address_to_script(address)
            self.assertEqual(
                address,
                bitcoin.script_to_address(scriptPubKey, self.test_signet_params[0])
            )

    def test_bf_wif_priv(self):
        for _ in range(0, 10000):
            if not _ % 1000:
                print(_)
            priv = sha256(os.urandom(32))
            wif = bitcoin.encode_privkey(priv, 'wif', vbyte=self.test_signet_params[2])
            self.assertTrue(wif.startswith('8'), msg=wif)
            self.assertEqual(bitcoin.encode_privkey(wif, 'hex'), priv)



from bitcoin.py2specials import *
from bitcoin.py3specials import *
from bitcoin.main import *
from bitcoin.transaction import *
from bitcoin.deterministic import *
from bitcoin.bci import *
from bitcoin.composite import *
from bitcoin.stealth import *
from bitcoin.blocks import *
from bitcoin.mnemonic import *
from bitcoin.segwit import *
from bitcoin.bech32 import *



# Majority of code taken from reference implementation:
# https://github.com/sipa/bech32/blob/master/ref/python/segwit_addr.py

# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Reference implementation for Bech32 and segwit addresses."""
import binascii
import re

from bitcoin import encode_pubkey, hash160, SIGHASH_ALL, privkey_to_pubkey, encode, hashlib

from enum import Enum

class Encoding(Enum):
    """Enumeration type to list the various supported encodings."""
    BECH32 = 1
    BECH32M = 2

CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"
BECH32M_CONST = 0x2bc830a3

BECH32_BITCOIN_PREFIX = 'bc'
BECH32_BITCOIN_TESTNET_PREFIX = 'tb'
BECH32_BITCOIN_REGTEST_PREFIX = 'bcrt'

BECH32_CUSTOM_SIGNET_PREFIX = 'cbc'
BECH32_CUSTOM_SIGNET_TESTNET_PREFIX = 'cbt'

BECH32_PREFIXES = [
    BECH32_CUSTOM_SIGNET_TESTNET_PREFIX,
    BECH32_CUSTOM_SIGNET_PREFIX,
    BECH32_BITCOIN_PREFIX,
    BECH32_BITCOIN_TESTNET_PREFIX,
    BECH32_BITCOIN_REGTEST_PREFIX
]


def bech32_polymod(values):
    """Internal function that computes the Bech32 checksum."""
    generator = [0x3b6a57b2, 0x26508e6d, 0x1ea119fa, 0x3d4233dd, 0x2a1462b3]
    chk = 1
    for value in values:
        top = chk >> 25
        chk = (chk & 0x1ffffff) << 5 ^ value
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def bech32_hrp_expand(hrp):
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]

'''
def bech32_verify_checksum(hrp, data):
    """Verify a checksum given HRP and converted data characters."""
    return bech32_polymod(bech32_hrp_expand(hrp) + data) == 1
'''

def bech32_verify_checksum(hrp, data):
    """Verify a checksum given HRP and converted data characters."""
    const = bech32_polymod(bech32_hrp_expand(hrp) + data)
    if const == 1:
        return Encoding.BECH32
    if const == BECH32M_CONST:
        return Encoding.BECH32M
    return None

'''
def bech32_create_checksum(hrp, data):
    """Compute the checksum values given HRP and data."""
    values = bech32_hrp_expand(hrp) + data
    polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]
'''

def bech32_create_checksum(hrp, data, spec):
    """Compute the checksum values given HRP and data."""
    values = bech32_hrp_expand(hrp) + data
    const = BECH32M_CONST if spec == Encoding.BECH32M else 1
    polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ const
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]


def bech32_encode(hrp, data, spec):
    """Compute a Bech32 string given HRP and data values."""
    combined = data + bech32_create_checksum(hrp, data, spec)
    return hrp + '1' + ''.join([CHARSET[d] for d in combined])

'''
def bech32_decode(bech):
    """Validate a Bech32 string, and determine HRP and data."""
    if ((any(ord(x) < 33 or ord(x) > 126 for x in bech)) or
            (bech.lower() != bech and bech.upper() != bech)):
        return None, None
    bech = bech.lower()
    pos = bech.rfind('1')
    if pos < 1 or pos + 7 > len(bech) or len(bech) > 90:
        return None, None
    if not all(x in CHARSET for x in bech[pos+1:]):
        return None, None
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos+1:]]
    if not bech32_verify_checksum(hrp, data):
        return None, None
    return hrp, data[:-6]
'''

def _bech32_decode(bech):
    """Validate a Bech32/Bech32m string, and determine HRP and data."""
    if ((any(ord(x) < 33 or ord(x) > 126 for x in bech)) or
            (bech.lower() != bech and bech.upper() != bech)):
        return (None, None, None)
    bech = bech.lower()
    pos = bech.rfind('1')
    if pos < 1 or pos + 7 > len(bech) or len(bech) > 90:
        return (None, None, None)
    if not all(x in CHARSET for x in bech[pos+1:]):
        return (None, None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos+1:]]
    spec = bech32_verify_checksum(hrp, data)
    if spec is None:
        return (None, None, None)
    return (hrp, data[:-6], spec)


def convertbits(data, frombits, tobits, pad=True):
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


def _decode(hrp, addr):
    """Decode a segwit address."""
    hrpgot, data, spec = _bech32_decode(addr)
    if hrpgot != hrp:
        return (None, None)
    decoded = convertbits(data[1:], 5, 8, False)
    if decoded is None or len(decoded) < 2 or len(decoded) > 40:
        return (None, None)
    if data[0] > 16:
        return (None, None)
    if data[0] == 0 and len(decoded) != 20 and len(decoded) != 32:
        return (None, None)
    if data[0] == 0 and spec != Encoding.BECH32 or data[0] != 0 and spec != Encoding.BECH32M:
        return (None, None)
    return (data[0], decoded)


def _bech32_encode(hrp, data, spec):
    """Compute a Bech32 string given HRP and data values."""
    combined = data + bech32_create_checksum(hrp, data, spec)
    return hrp + '1' + ''.join([CHARSET[d] for d in combined])


def _encode(hrp, witver, witprog):
    """Encode a segwit address."""
    spec = Encoding.BECH32 if witver == 0 else Encoding.BECH32M
    ret = _bech32_encode(hrp, [witver] + convertbits(witprog, 8, 5), spec)
    if _decode(hrp, ret) == (None, None):
        return None
    return ret


def bech32encode(script: str, prefix=BECH32_BITCOIN_PREFIX):
    from bitcoin.transaction import deserialize_script
    witnessversion, witnessprogram = deserialize_script(script)
    witnessversion = 0 if witnessversion is None else witnessversion
    assert 0 <= witnessversion <= 16
    if witnessversion == 0:
        assert len(witnessprogram) == 40 or len(witnessprogram) == 64
    return _encode(prefix, witnessversion, bytearray(binascii.unhexlify(witnessprogram))).lower()


def int_to_hex(n: int) -> str:
    return binascii.hexlify(chr(n).encode()).decode()


def segwit_scriptpubkey(witver, witprog):
    """Construct a Segwit scriptPubKey for a given witness program."""
    return bytes([witver + 0x50 if witver else 0, len(witprog)] + witprog)


def bech32decode(text: str):
    assert isinstance(text, str)
    from bitcoin import from_int_to_byte
    if text[:4].lower() == BECH32_BITCOIN_REGTEST_PREFIX:
        human_readable_part_size = 4
    elif text[:2].lower() in (
        BECH32_BITCOIN_PREFIX,
        BECH32_BITCOIN_TESTNET_PREFIX
    ):
        human_readable_part_size = 2
    elif text[:3].lower() in (
        BECH32_CUSTOM_SIGNET_PREFIX,
        BECH32_CUSTOM_SIGNET_TESTNET_PREFIX
    ):
        human_readable_part_size = 3
    else:
        raise ValueError('Invalid readable part')

    if len(set([t.islower() for t in text if t.isalpha()])) > 1:
        raise ValueError('Mixed case')
    text = text.lower().encode("utf-8").decode()

    witnessversion, wit_ordnallist = _decode(text[:human_readable_part_size], text)
    if witnessversion is None and wit_ordnallist is None:
        raise ValueError('Exception with %s (prefix: %s)' % (text, text[:human_readable_part_size]))
    if not witnessversion:
        assert len(wit_ordnallist) == 20 or len(wit_ordnallist) == 32
    o = ""
    for ordinal in wit_ordnallist:
        o += binascii.hexlify(from_int_to_byte(ordinal)).decode()
    return int_to_hex(witnessversion + 0x50 if witnessversion else 0) + int_to_hex(int(len(o) / 2)) + o


def pubkey_to_bech32_address(pubkey, prefix=BECH32_BITCOIN_PREFIX):
    if isinstance(pubkey, (list, tuple)):
        pubkey = encode_pubkey(pubkey, 'bin')
    if len(pubkey) in [66, 130]:
        pubkey = binascii.unhexlify(pubkey)
        pubkey_hash = hash160(pubkey)
        return bech32encode('0014' + pubkey_hash, prefix=prefix)
    if len(pubkey) in [68]:
        return bech32encode(pubkey, prefix=prefix)
    raise ValueError()


def bech32_sign(tx, i, priv, amount, script=None, hashcode=SIGHASH_ALL):
    from bitcoin import deserialize, segwit_signature_form, ecdsa_raw_sign, der_encode_sig, serialize, compress
    i = int(i)
    txobj = tx if isinstance(tx, dict) else deserialize(tx)
    if len(priv) <= 33:
        priv = binascii.hexlify(priv)
    pub = compress(privkey_to_pubkey(priv))
    script = script or ('76a914' + hash160(binascii.unhexlify(pub)) + '88ac')
    signing_tx = segwit_signature_form(tx, i, script, amount, hashcode=hashcode)
    rawsig = ecdsa_raw_sign(hashlib.sha256(hashlib.sha256(binascii.unhexlify(signing_tx)).digest()).hexdigest(), priv)
    sig = der_encode_sig(*rawsig)+encode(hashcode, 16, 2)
    txobj['ins'][i]['txinwitness'] = [sig, pub]
    return serialize(txobj)


def bech32_script_to_address(script, prefix=BECH32_BITCOIN_PREFIX):
    from bitcoin import sha256
    script = binascii.unhexlify(script)
    script_hash = sha256(script)
    return bech32encode('0020' + script_hash, prefix=prefix)


def bech32_scripthash_to_address(scripthash: str, prefix=BECH32_BITCOIN_PREFIX):
    if re.match('^[0-9a-fA-F]*$', scripthash):
        scripthash = bytes.fromhex(scripthash)

    if scripthash[:2] == b'\x00\x20':
        return bech32encode(scripthash.hex(), prefix)
    if scripthash[:2] == b'\x00\x14':
        return bech32encode(scripthash.hex(), prefix)
    raise ValueError

def bech32_multisign(tx, i, priv, amount, script, hashcode=SIGHASH_ALL):
    from bitcoin import segwit_signature_form, ecdsa_raw_sign, der_encode_sig
    i = int(i)
    if len(priv) <= 33:
        priv = binascii.hexlify(priv)
    signing_tx = segwit_signature_form(tx, i, script, amount, hashcode=hashcode)
    rawsig = ecdsa_raw_sign(hashlib.sha256(hashlib.sha256(binascii.unhexlify(signing_tx)).digest()).hexdigest(), priv)
    sig = der_encode_sig(*rawsig)+encode(hashcode, 16, 2)
    return sig


def apply_bech32_multisignatures(tx, i, witness_program, signatures):
    from bitcoin import deserialize, serialize
    o = [""] + signatures + [witness_program]
    txobj = deserialize(tx)
    txobj['ins'][i]['txinwitness'] = o
    return serialize(txobj)



import unittest
from bitcoin import *

class TestTransaction(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Attempting transaction creation")

    def test3(self):
        print(
            deserialize_script('52210248905f94419795ea33cd42474e10bfaddc3ee5f0f0c66ecc29238fea6555f29c2103fde505b2f67b2c8ec17c7540bbc9aafb527366c0863d655d03a00e5f3c4bbbd121023f96141f1bec4df22465539ecd807762e2c96b75e436540d3e7654d461b62a1953ae')
        )

    def test2(self):
        pub = '029b06d73294a2fe59dd5d2156f9d7bf1cadc8e741b39fff834d39a055ab8f5c97'
        addr = 'bcrt1q8s2hkukgulyf575hakxazset8v2z5ltxnvepy8'
        self.assertEqual(pubkey_to_bech32_address(pub, prefix='bcrt'), addr)
        print(deserialize_script('00141976a9141d0f172a0ecb48aee1be1f2687d2963ae33f71a188ac'))
        print(hash160(binascii.unhexlify('025476c2e83188368da1ff3e292e7acafcdb3566bb0ad253f62fc70f07aeee6357')))

    def test(self):
        addr = 'bcrt1qpkye5029qngu6jflsg4wnmrtmj47py4mhsvhav'
        priv = sha256(b'priv_key_text')
        pub = compress(privtopub(priv))
        address = pubkey_to_bech32_address(pub, prefix='bcrt')
        self.assertEqual(addr, address)
        transaction_to_sign = mktx(
            {
                'output': '2ecfbf58dd128a3f1a6d4c1cdb1740c706389494c7ed94b9e027582d71602ade:1',
                'segregated': True
            },
            [
                {'address': addr, 'value': int(1 * 10**8) - 5000}
            ]
        )
        tx = bech32_sign(transaction_to_sign, 0, priv, int(1 * 10**8))
        self.assertEqual(segwit_txhash(tx)['txid'], '2fe3694f5b62097211a04521593eaa9661ff6beb5aec9696efc2ce6d2b78959d')

    def test_multisig(self):
        priv1 = sha256(b'priv_key_text')
        priv2 = sha256(b'priv_key_text_2')
        priv3 = sha256(b'priv_key_text_3')
        pub1 = compress(privtopub(priv1))
        pub2 = compress(privtopub(priv2))
        pub3 = compress(privtopub(priv3))
        witness_program = mk_multisig_script([pub1, pub2, pub3], 2, 3)
        addr = bech32_script_to_address(witness_program, prefix='bcrt')
        self.assertEqual(
            witness_program,
            '52210248905f94419795ea33cd42474e10bfaddc3ee5f0f0c66ecc'
            '29238fea6555f29c2103fde505b2f67b2c8ec17c7540bbc9aafb52'
            '7366c0863d655d03a00e5f3c4bbbd121023f96141f1bec4df22465'
            '539ecd807762e2c96b75e436540d3e7654d461b62a1953ae'
        )
        self.assertEqual(
            addr, 'bcrt1qd3rdg790qea4gdw0k9lmp5tjl74q0hg9t09x64x2tnt9zl0sepasqjv7vn'
        )
        transaction_to_sign = mktx(
            {
                'output': '940fe2aa34c4d29b2df0e0486d4cb7e6821f426136d87e7a1a949d43ad248340:0',
                'segregated': True
            },
            [
                {'address': addr, 'value': int(1 * 10**8) - 5000}
            ]
        )
        sig1 = bech32_multisign(transaction_to_sign, 0, priv1, int(1 * 10 ** 8), witness_program)
        sig2 = bech32_multisign(transaction_to_sign, 0, priv2, int(1 * 10 ** 8), witness_program)
        self.assertEqual(
            apply_bech32_multisignatures(transaction_to_sign, 0, witness_program, [sig1, sig2]),
            "01000000000101408324ad439d941a7a7ed83661421f82e6b74c6d48e0f02d9bd2c434aae20f94000000"
            "0000ffffffff0178cdf505000000002200206c46d478af067b5435cfb17fb0d172ffaa07dd055bca6d54"
            "ca5cd6517df0c87b040048304502210092e960ab3ffe263620e22a4b5b61b452bbeee04aae736e237a09"
            "47027642a6310220116cc4db1afd7ed02dc27caea2c01cd578ad9c1b4b5531e8af129f66aa75b2a60147"
            "30440220129f3bdc8bd5a62b24cd32cb920703961ffde897e430fe3f3b47141dc5852731022077b869a8"
            "c33b7d016fb96b23c3bfffa638cc4a7ca87c4129218183ae0ed45c44016952210248905f94419795ea33"
            "cd42474e10bfaddc3ee5f0f0c66ecc29238fea6555f29c2103fde505b2f67b2c8ec17c7540bbc9aafb52"
            "7366c0863d655d03a00e5f3c4bbbd121023f96141f1bec4df22465539ecd807762e2c96b75e436540d3e"
            "7654d461b62a1953ae00000000"
        )

    def test_signet(self):
        priv = 'bLnPotUnS3sQRefhjFPFNcJ5WS92ajuGtpsrGrqcu8reweQCiJpt'
        privhex = encode_privkey(decode_privkey(priv), 'hex', vbyte=0x6f)
        pub = compress(privtopub(privhex))
        addr = pubkey_to_bech32_address(pub, prefix='cbt')
        print(addr)
        print(pub)
        transaction_to_sign = mktx(
            [{
                'output': 'aef21e0003b31fc2cb8bafa83c271102b7ea06a5e70ac10deedc05fa86c0d697:0',
                'segregated': True
            }],
            [
                {'address': addr, 'value': int(200000 * 10**8)}
            ]
        )
        tx = bech32_sign(transaction_to_sign, 0, privhex, int(200000 * 10**8))
        print(tx)

    def test_oldscriptoaddr(self):
        scripts = {
            '00201863143c14c5166804bd19203356da136c985678cd4d27a1b8c6329604903262': [
                'tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7',
                196, False, BECH32_BITCOIN_TESTNET_PREFIX
            ],
            '00202a11fadf4a96c60669ae50b2ac7c0b3cae0b3d94483670504dc154a255826322': [
              'bcrt1q9ggl4h62jmrqv6dw2ze2clqt8jhqk0v5fqm8q5zdc922y4vzvv3qj3le86',
                196, True, BECH32_BITCOIN_REGTEST_PREFIX
            ],
            '00142a3af484ba735f3de4535f0fd522d4666646ef3f': [
                'bcrt1q9ga0fp96wd0nmezntu8a2gk5venydmelf8ddtm',
                111, True, BECH32_BITCOIN_REGTEST_PREFIX
            ],
            '00143262354a6825b39de7f35534ddb04f45ea7a1f43': [
                'bc1qxf3r2jngykeemeln256dmvz0gh48586rfx4j2p',
                0, False, BECH32_BITCOIN_PREFIX
            ]
        }

        for script in scripts:
            resp = scripthash_to_address(script, scripts[script][1], 0, scripts[script][3])
            self.assertEqual(scripts[script][0], resp)
            b32_resp = bech32_scripthash_to_address(script, prefix=scripts[script][3])
            self.assertEqual(b32_resp, resp)

    def test_multisig_upper(self):
        priv1 = sha256(b'priv_key_text')
        priv2 = sha256(b'priv_key_text_2')
        priv3 = sha256(b'priv_key_text_3')
        pub1 = compress(privtopub(priv1))
        pub2 = compress(privtopub(priv2))
        pub3 = compress(privtopub(priv3))
        witness_program = mk_multisig_script([pub1, pub2, pub3], 2, 3)
        addr = bech32_script_to_address(witness_program, prefix='bcrt')
        self.assertEqual(
            witness_program,
            '52210248905f94419795ea33cd42474e10bfaddc3ee5f0f0c66ecc'
            '29238fea6555f29c2103fde505b2f67b2c8ec17c7540bbc9aafb52'
            '7366c0863d655d03a00e5f3c4bbbd121023f96141f1bec4df22465'
            '539ecd807762e2c96b75e436540d3e7654d461b62a1953ae'
        )
        self.assertEqual(
            addr, 'bcrt1qd3rdg790qea4gdw0k9lmp5tjl74q0hg9t09x64x2tnt9zl0sepasqjv7vn'
        )
        transaction_to_sign = mktx(
            {
                'output': '940fe2aa34c4d29b2df0e0486d4cb7e6821f426136d87e7a1a949d43ad248340:0',
                'segregated': True
            },
            [
                {'address': addr.upper(), 'value': int(1 * 10**8) - 5000}
            ]
        )
        sig1 = bech32_multisign(transaction_to_sign, 0, priv1, int(1 * 10 ** 8), witness_program)
        sig2 = bech32_multisign(transaction_to_sign, 0, priv2, int(1 * 10 ** 8), witness_program)
        self.assertEqual(
            apply_bech32_multisignatures(transaction_to_sign, 0, witness_program, [sig1, sig2]),
            "01000000000101408324ad439d941a7a7ed83661421f82e6b74c6d48e0f02d9bd2c434aae20f94000000"
            "0000ffffffff0178cdf505000000002200206c46d478af067b5435cfb17fb0d172ffaa07dd055bca6d54"
            "ca5cd6517df0c87b040048304502210092e960ab3ffe263620e22a4b5b61b452bbeee04aae736e237a09"
            "47027642a6310220116cc4db1afd7ed02dc27caea2c01cd578ad9c1b4b5531e8af129f66aa75b2a60147"
            "30440220129f3bdc8bd5a62b24cd32cb920703961ffde897e430fe3f3b47141dc5852731022077b869a8"
            "c33b7d016fb96b23c3bfffa638cc4a7ca87c4129218183ae0ed45c44016952210248905f94419795ea33"
            "cd42474e10bfaddc3ee5f0f0c66ecc29238fea6555f29c2103fde505b2f67b2c8ec17c7540bbc9aafb52"
            "7366c0863d655d03a00e5f3c4bbbd121023f96141f1bec4df22465539ecd807762e2c96b75e436540d3e"
            "7654d461b62a1953ae00000000"
        )


import unittest

import bitcoin
from bitcoin import *


class TestTransaction(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Attempting transaction creation")

    def test3(self):
        print(
            deserialize_script('52210248905f94419795ea33cd42474e10bfaddc3ee5f0f0c66ecc29238fea6555f29c2103fde505b2f67b2c8ec17c7540bbc9aafb527366c0863d655d03a00e5f3c4bbbd121023f96141f1bec4df22465539ecd807762e2c96b75e436540d3e7654d461b62a1953ae')
        )

    def test2(self):
        pub = '029b06d73294a2fe59dd5d2156f9d7bf1cadc8e741b39fff834d39a055ab8f5c97'
        addr = 'bcrt1q8s2hkukgulyf575hakxazset8v2z5ltxnvepy8'
        self.assertEqual(pubkey_to_bech32_address(pub, prefix='bcrt'), addr)
        print(deserialize_script('00141976a9141d0f172a0ecb48aee1be1f2687d2963ae33f71a188ac'))
        print(hash160(binascii.unhexlify('025476c2e83188368da1ff3e292e7acafcdb3566bb0ad253f62fc70f07aeee6357')))

    def test_multisig(self):
        priv1 = sha256(b'sighash_priv_key_text')
        priv2 = sha256(b'sighash_priv_key_text_2')
        pub1 = compress(privtopub(priv1))
        pub2 = compress(privtopub(priv2))
        witness_program = mk_multisig_script([pub1, pub2], 2, 2)
        addr = bech32_script_to_address(witness_program, prefix='bc')
        print('addr', addr)
        recipient = '3AbjFnwcChgaAGsPx28hnpDWF3yUobvTFT'
        amount = 0.00028295
        transaction_to_sign = mktx(
            {
                'output': '99911f6ddabc51290a45194f268d7e618284d7f42d79a2b57bee9bc5b11787c5:0',
                'segregated': True
            },
            [
                {'address': recipient, 'value': int(amount * 10**8) - 4500}
            ]
        )

        tx = bitcoin.deserialize(transaction_to_sign)
        """ test big opreturn size"""
        bigscript = [os.urandom(1024).hex() for _ in range(0, 1000)]
        tx['outs'].append(
            {
                'value': 0,
                'script': '00' + bitcoin.serialize_script(bigscript)
            }
        )
        txs = bitcoin.serialize(tx)
        tx = bitcoin.deserialize(txs)
        s = bitcoin.deserialize_script(tx['outs'][-1]['script'])
        self.assertEqual(s[0], None)
        self.assertEqual(s[1:], bigscript)

        sig1 = bech32_multisign(
            transaction_to_sign, 0, priv1, int(amount * 10 ** 8),
            witness_program, hashcode=SIGHASH_NONE|SIGHASH_ANYONECANPAY
        )
        sig2 = bech32_multisign(transaction_to_sign, 0, priv2, int(amount * 10 ** 8), witness_program)
        tx = apply_bech32_multisignatures(transaction_to_sign, 0, witness_program, [sig1, sig2])
        print(tx)

    def test_hash_opreturn(self):
        tx = '0100000000010122371ebb7a0432f0d506c35c8a78da70d29258dd50fc870426b3ced80839ebe50100000000fdffffff03983a00000000000017a9148380f47f331682e3683cc0628b04d3e1c918af8887464d00000000000017a914cc2008ff35eea6390b32dde0cf5998fd1016fcec8700000000000000005100160014636f6e696f5f66726f7a656e5f6f75747075747301010102040100000020e5eb3908d8ceb3260487fc50dd5892d270da788a5cc306d5f032047abb1e372202010008de8700000000000002020002000000000000'
        txhash = bitcoin.segwit_txhash(tx)
        print(txhash)
        des_tx = bitcoin.deserialize(tx)
        des_tx['outs'] = des_tx['outs'][:-1]
        tx2 = bitcoin.serialize(des_tx)
        txhash = bitcoin.segwit_txhash(tx2)
        print(txhash)



#!/usr/bin/python
import json, re
import random
import sys
import binascii
from bitcoin import safe_hexlify

try:
    from urllib.request import build_opener
except:
    from urllib2 import build_opener


# Makes a request to a given URL (first arg) and optional params (second arg)
def make_request(*args):
    opener = build_opener()
    opener.addheaders = [('User-agent',
                          'Mozilla/5.0'+str(random.randrange(1000000)))]
    try:
        return opener.open(*args).read().strip()
    except Exception as e:
        try:
            p = e.read().strip()
        except:
            p = e
        raise Exception(p)


def is_testnet(inp):
    '''Checks if inp is a testnet address or if UTXO is a known testnet TxID''' 
    if isinstance(inp, (list, tuple)) and len(inp) >= 1:
        return any([is_testnet(x) for x in inp])
    elif not isinstance(inp, str):    # sanity check
        raise TypeError("Input must be str/unicode, not type %s" % str(type(inp)))

    if not inp or (inp.lower() in ("btc", "testnet")): 
        pass

    ## ADDRESSES
    if inp[0] in "123mn":
        if re.match("^[2mn][a-km-zA-HJ-NP-Z0-9]{26,33}$", inp):
            return True
        elif re.match("^[13][a-km-zA-HJ-NP-Z0-9]{26,33}$", inp):
            return False
        else:
            #sys.stderr.write("Bad address format %s")
            return None

    ## TXID
    elif re.match('^[0-9a-fA-F]{64}$', inp):
        base_url = "http://api.blockcypher.com/v1/btc/{network}/txs/{txid}?includesHex=false"
        try:
            # try testnet fetchtx
            make_request(base_url.format(network="test3", txid=inp.lower()))
            return True
        except:
            # try mainnet fetchtx
            make_request(base_url.format(network="main", txid=inp.lower()))
            return False
        sys.stderr.write("TxID %s has no match for testnet or mainnet (Bad TxID)")
        return None
    else:
        raise TypeError("{0} is unknown input".format(inp))


def set_network(*args):
    '''Decides if args for unspent/fetchtx/pushtx are mainnet or testnet'''
    r = []
    for arg in args:
        if not arg: 
            pass
        if isinstance(arg, str):
            r.append(is_testnet(arg))
        elif isinstance(arg, (list, tuple)):
            return set_network(*arg)
    if any(r) and not all(r):
        raise Exception("Mixed Testnet/Mainnet queries")
    return "testnet" if any(r) else "btc"


def parse_addr_args(*args):
    # Valid input formats: unspent([addr1, addr2, addr3])
    #                      unspent([addr1, addr2, addr3], network)
    #                      unspent(addr1, addr2, addr3)
    #                      unspent(addr1, addr2, addr3, network)
    addr_args = args
    network = "btc"
    if len(args) == 0:
        return [], 'btc'
    if len(args) >= 1 and args[-1] in ('testnet', 'btc'):
        network = args[-1]
        addr_args = args[:-1]
    if len(addr_args) == 1 and isinstance(addr_args, list):
        network = set_network(*addr_args[0])
        addr_args = addr_args[0]
    if addr_args and isinstance(addr_args, tuple) and isinstance(addr_args[0], list):
        addr_args = addr_args[0]
    network = set_network(addr_args)
    return addr_args, network   # note params are "reversed" now


# Gets the unspent outputs of one or more addresses
def bci_unspent(*args):
    addrs, network = parse_addr_args(*args)
    u = []
    for a in addrs:
        try:
            data = make_request('https://blockchain.info/unspent?active='+a)
        except Exception as e:
            if str(e) == 'No free outputs to spend':
                continue
            else:
                raise Exception(e)
        jsonobj = json.loads(data.decode("utf-8"))
        for o in jsonobj["unspent_outputs"]:
            h = safe_hexlify(binascii.unhexlify(o['tx_hash'])[::-1])
            u.append({
                "output": h+':'+str(o['tx_output_n']),
                "value": o['value']
            })
    return u


def blockr_unspent(*args):
    # Valid input formats: blockr_unspent([addr1, addr2,addr3])
    #                      blockr_unspent(addr1, addr2, addr3)
    #                      blockr_unspent([addr1, addr2, addr3], network)
    #                      blockr_unspent(addr1, addr2, addr3, network)
    # Where network is 'btc' or 'testnet'
    network, addr_args = parse_addr_args(*args)

    if network == 'testnet':
        blockr_url = 'http://tbtc.blockr.io/api/v1/address/unspent/'
    elif network == 'btc':
        blockr_url = 'http://btc.blockr.io/api/v1/address/unspent/'
    else:
        raise Exception(
            'Unsupported network {0} for blockr_unspent'.format(network))

    if len(addr_args) == 0:
        return []
    elif isinstance(addr_args[0], list):
        addrs = addr_args[0]
    else:
        addrs = addr_args
    res = make_request(blockr_url+','.join(addrs))
    try:
        data = json.loads(res.decode('utf-8'))['data']
    except:
        data = json.loads(res)['data']

    o = []
    if 'unspent' in data:
        data = [data]
    for dat in data:
        for u in dat['unspent']:
            o.append({
                "output": u['tx']+':'+str(u['n']),
                "value": int(u['amount'].replace('.', ''))
            })
    return o


def helloblock_unspent(*args):
    addrs, network = parse_addr_args(*args)
    if network == 'testnet':
        url = 'https://testnet.helloblock.io/v1/addresses/%s/unspents?limit=500&offset=%s'
    elif network == 'btc':
        url = 'https://mainnet.helloblock.io/v1/addresses/%s/unspents?limit=500&offset=%s'
    o = []
    for addr in addrs:
        for offset in xrange(0, 10**9, 500):
            res = make_request(url % (addr, offset))
            data = json.loads(res.decode("utf-8"))["data"]
            if not len(data["unspents"]):
                break
            elif offset:
                sys.stderr.write("Getting more unspents: %d\n" % offset)
            for dat in data["unspents"]:
                o.append({
                    "output": dat["txHash"]+':'+str(dat["index"]),
                    "value": dat["value"],
                })
    return o


unspent_getters = {
    'bci': bci_unspent,
    'blockr': blockr_unspent,
    'helloblock': helloblock_unspent
}


def unspent(*args, **kwargs):
    f = unspent_getters.get(kwargs.get('source', ''), bci_unspent)
    return f(*args)


# Gets the transaction output history of a given set of addresses,
# including whether or not they have been spent
def history(*args):
    # Valid input formats: history([addr1, addr2,addr3])
    #                      history(addr1, addr2, addr3)
    if len(args) == 0:
        return []
    elif isinstance(args[0], list):
        addrs = args[0]
    else:
        addrs = args

    txs = []
    for addr in addrs:
        offset = 0
        while 1:
            gathered = False
            while not gathered:
                try:
                    data = make_request(
                        'https://blockchain.info/address/%s?format=json&offset=%s' %
                        (addr, offset))
                    gathered = True
                except Exception as e:
                    try:
                        sys.stderr.write(e.read().strip())
                    except:
                        sys.stderr.write(str(e))
                    gathered = False
            try:
                jsonobj = json.loads(data.decode("utf-8"))
            except:
                raise Exception("Failed to decode data: "+data)
            txs.extend(jsonobj["txs"])
            if len(jsonobj["txs"]) < 50:
                break
            offset += 50
            sys.stderr.write("Fetching more transactions... "+str(offset)+'\n')
    outs = {}
    for tx in txs:
        for o in tx["out"]:
            if o.get('addr', None) in addrs:
                key = str(tx["tx_index"])+':'+str(o["n"])
                outs[key] = {
                    "address": o["addr"],
                    "value": o["value"],
                    "output": tx["hash"]+':'+str(o["n"]),
                    "block_height": tx.get("block_height", None)
                }
    for tx in txs:
        for i, inp in enumerate(tx["inputs"]):
            if "prev_out" in inp:
                if inp["prev_out"].get("addr", None) in addrs:
                    key = str(inp["prev_out"]["tx_index"]) + \
                        ':'+str(inp["prev_out"]["n"])
                    if outs.get(key):
                        outs[key]["spend"] = tx["hash"]+':'+str(i)
    return [outs[k] for k in outs]


# Pushes a transaction to the network using https://blockchain.info/pushtx
def bci_pushtx(tx):
    if not re.match('^[0-9a-fA-F]*$', tx):
        tx = tx.encode('hex')
    return make_request('https://blockchain.info/pushtx', 'tx='+tx)


def eligius_pushtx(tx):
    if not re.match('^[0-9a-fA-F]*$', tx):
        tx = tx.encode('hex')
    s = make_request(
        'http://eligius.st/~wizkid057/newstats/pushtxn.php',
        'transaction='+tx+'&send=Push')
    strings = re.findall('string[^"]*"[^"]*"', s)
    for string in strings:
        quote = re.findall('"[^"]*"', string)[0]
        if len(quote) >= 5:
            return quote[1:-1]


def blockr_pushtx(tx, network='btc'):
    if network == 'testnet':
        blockr_url = 'http://tbtc.blockr.io/api/v1/tx/push'
    elif network == 'btc':
        blockr_url = 'http://btc.blockr.io/api/v1/tx/push'
    else:
        raise Exception(
            'Unsupported network {0} for blockr_pushtx'.format(network))

    if not re.match('^[0-9a-fA-F]*$', tx):
        tx = tx.encode('hex')
    return make_request(blockr_url, '{"hex":"%s"}' % tx)


def helloblock_pushtx(tx):
    if not re.match('^[0-9a-fA-F]*$', tx):
        tx = tx.encode('hex')
    return make_request('https://mainnet.helloblock.io/v1/transactions',
                        'rawTxHex='+tx)

pushtx_getters = {
    'bci': bci_pushtx,
    'blockr': blockr_pushtx,
    'helloblock': helloblock_pushtx
}


def pushtx(*args, **kwargs):
    f = pushtx_getters.get(kwargs.get('source', ''), bci_pushtx)
    return f(*args)


def last_block_height(network='btc'):
    if network == 'testnet':
        data = make_request('http://tbtc.blockr.io/api/v1/block/info/last')
        jsonobj = json.loads(data.decode("utf-8"))
        return jsonobj["data"]["nb"]

    data = make_request('https://blockchain.info/latestblock')
    jsonobj = json.loads(data.decode("utf-8"))
    return jsonobj["height"]


# Gets a specific transaction
def bci_fetchtx(txhash):
    if isinstance(txhash, list):
        return [bci_fetchtx(h) for h in txhash]
    if not re.match('^[0-9a-fA-F]*$', txhash):
        txhash = txhash.encode('hex')
    data = make_request('https://blockchain.info/rawtx/'+txhash+'?format=hex')
    return data


def blockr_fetchtx(txhash, network='btc'):
    if network == 'testnet':
        blockr_url = 'http://tbtc.blockr.io/api/v1/tx/raw/'
    elif network == 'btc':
        blockr_url = 'http://btc.blockr.io/api/v1/tx/raw/'
    else:
        raise Exception(
            'Unsupported network {0} for blockr_fetchtx'.format(network))
    if isinstance(txhash, list):
        txhash = ','.join([x.encode('hex') if not re.match('^[0-9a-fA-F]*$', x)
                           else x for x in txhash])
        jsondata = json.loads(make_request(blockr_url+txhash).decode("utf-8"))
        return [d['tx']['hex'] for d in jsondata['data']]
    else:
        if not re.match('^[0-9a-fA-F]*$', txhash):
            txhash = txhash.encode('hex')
        jsondata = json.loads(make_request(blockr_url+txhash).decode("utf-8"))
        return jsondata['data']['tx']['hex']


def helloblock_fetchtx(txhash, network='btc'):
    if isinstance(txhash, list):
        return [helloblock_fetchtx(h) for h in txhash]
    if not re.match('^[0-9a-fA-F]*$', txhash):
        txhash = txhash.encode('hex')
    if network == 'testnet':
        url = 'https://testnet.helloblock.io/v1/transactions/'
    elif network == 'btc':
        url = 'https://mainnet.helloblock.io/v1/transactions/'
    else:
        raise Exception(
            'Unsupported network {0} for helloblock_fetchtx'.format(network))
    data = json.loads(make_request(url + txhash).decode("utf-8"))["data"]["transaction"]
    o = {
        "locktime": data["locktime"],
        "version": data["version"],
        "ins": [],
        "outs": []
    }
    for inp in data["inputs"]:
        o["ins"].append({
            "script": inp["scriptSig"],
            "outpoint": {
                "index": inp["prevTxoutIndex"],
                "hash": inp["prevTxHash"],
            },
            "sequence": 4294967295
        })
    for outp in data["outputs"]:
        o["outs"].append({
            "value": outp["value"],
            "script": outp["scriptPubKey"]
        })
    from bitcoin.transaction import serialize
    from bitcoin.transaction import txhash as TXHASH
    tx = serialize(o)
    assert TXHASH(tx) == txhash
    return tx


fetchtx_getters = {
    'bci': bci_fetchtx,
    'blockr': blockr_fetchtx,
    'helloblock': helloblock_fetchtx
}


def fetchtx(*args, **kwargs):
    f = fetchtx_getters.get(kwargs.get('source', ''), bci_fetchtx)
    return f(*args)


def firstbits(address):
    if len(address) >= 25:
        return make_request('https://blockchain.info/q/getfirstbits/'+address)
    else:
        return make_request(
            'https://blockchain.info/q/resolvefirstbits/'+address)


def get_block_at_height(height):
    j = json.loads(make_request("https://blockchain.info/block-height/" +
                   str(height)+"?format=json").decode("utf-8"))
    for b in j['blocks']:
        if b['main_chain'] is True:
            return b
    raise Exception("Block at this height not found")


def _get_block(inp):
    if len(str(inp)) < 64:
        return get_block_at_height(inp)
    else:
        return json.loads(make_request(
                          'https://blockchain.info/rawblock/'+inp).decode("utf-8"))


def bci_get_block_header_data(inp):
    j = _get_block(inp)
    return {
        'version': j['ver'],
        'hash': j['hash'],
        'prevhash': j['prev_block'],
        'timestamp': j['time'],
        'merkle_root': j['mrkl_root'],
        'bits': j['bits'],
        'nonce': j['nonce'],
    }

def blockr_get_block_header_data(height, network='btc'):
    if network == 'testnet':
        blockr_url = "http://tbtc.blockr.io/api/v1/block/raw/"
    elif network == 'btc':
        blockr_url = "http://btc.blockr.io/api/v1/block/raw/"
    else:
        raise Exception(
            'Unsupported network {0} for blockr_get_block_header_data'.format(network))

    k = json.loads(make_request(blockr_url + str(height)).decode("utf-8"))
    j = k['data']
    return {
        'version': j['version'],
        'hash': j['hash'],
        'prevhash': j['previousblockhash'],
        'timestamp': j['time'],
        'merkle_root': j['merkleroot'],
        'bits': int(j['bits'], 16),
        'nonce': j['nonce'],
    }


def get_block_timestamp(height, network='btc'):
    if network == 'testnet':
        blockr_url = "http://tbtc.blockr.io/api/v1/block/info/"
    elif network == 'btc':
        blockr_url = "http://btc.blockr.io/api/v1/block/info/"
    else:
        raise Exception(
            'Unsupported network {0} for get_block_timestamp'.format(network))

    import time, calendar
    if isinstance(height, list):
        k = json.loads(make_request(blockr_url + ','.join([str(x) for x in height])).decode("utf-8"))
        o = {x['nb']: calendar.timegm(time.strptime(x['time_utc'],
             "%Y-%m-%dT%H:%M:%SZ")) for x in k['data']}
        return [o[x] for x in height]
    else:
        k = json.loads(make_request(blockr_url + str(height)).decode("utf-8"))
        j = k['data']['time_utc']
        return calendar.timegm(time.strptime(j, "%Y-%m-%dT%H:%M:%SZ"))


block_header_data_getters = {
    'bci': bci_get_block_header_data,
    'blockr': blockr_get_block_header_data
}


def get_block_header_data(inp, **kwargs):
    f = block_header_data_getters.get(kwargs.get('source', ''),
                                      bci_get_block_header_data)
    return f(inp, **kwargs)


def get_txs_in_block(inp):
    j = _get_block(inp)
    hashes = [t['hash'] for t in j['tx']]
    return hashes


def get_block_height(txhash):
    j = json.loads(make_request('https://blockchain.info/rawtx/'+txhash).decode("utf-8"))
    return j['block_height']

# fromAddr, toAddr, 12345, changeAddress
def get_tx_composite(inputs, outputs, output_value, change_address=None, network=None):
    """mktx using blockcypher API"""
    inputs = [inputs] if not isinstance(inputs, list) else inputs
    outputs = [outputs] if not isinstance(outputs, list) else outputs
    network = set_network(change_address or inputs) if not network else network.lower()
    url = "http://api.blockcypher.com/v1/btc/{network}/txs/new?includeToSignTx=true".format(
                  network=('test3' if network=='testnet' else 'main'))
    is_address = lambda a: bool(re.match("^[123mn][a-km-zA-HJ-NP-Z0-9]{26,33}$", a))
    if any([is_address(x) for x in inputs]):
        inputs_type = 'addresses'        # also accepts UTXOs, only addresses supported presently
    if any([is_address(x) for x in outputs]):
        outputs_type = 'addresses'       # TODO: add UTXO support
    data = {
            'inputs':  [{inputs_type:  inputs}], 
            'confirmations': 0, 
            'preference': 'high', 
            'outputs': [{outputs_type: outputs, "value": output_value}]
            }
    if change_address:
        data["change_address"] = change_address    # 
    jdata = json.loads(make_request(url, data))
    hash, txh = jdata.get("tosign")[0], jdata.get("tosign_tx")[0]
    assert bin_dbl_sha256(txh.decode('hex')).encode('hex') == hash, "checksum mismatch %s" % hash
    return txh.encode("utf-8")

blockcypher_mktx = get_tx_composite



from bitcoin.main import *
from bitcoin.transaction import *
from bitcoin.bci import *
from bitcoin.deterministic import *
from bitcoin.blocks import *


# Takes privkey, address, value (satoshis), fee (satoshis)
def send(frm, to, value, fee=10000, **kwargs):
    return sendmultitx(frm, to + ":" + str(value), fee, **kwargs)


# Takes privkey, "address1:value1,address2:value2" (satoshis), fee (satoshis)
def sendmultitx(frm, *args, **kwargs):
    tv, fee = args[:-1], int(args[-1])
    outs = []
    outvalue = 0
    for a in tv:
        outs.append(a)
        outvalue += int(a.split(":")[1])

    u = unspent(privtoaddr(frm), **kwargs)
    u2 = select(u, int(outvalue)+int(fee))
    argz = u2 + outs + [privtoaddr(frm), fee]
    tx = mksend(*argz)
    tx2 = signall(tx, frm)
    return pushtx(tx2, **kwargs)


# Takes address, address, value (satoshis), fee(satoshis)
def preparetx(frm, to, value, fee=10000, **kwargs):
    tovalues = to + ":" + str(value)
    return preparemultitx(frm, tovalues, fee, **kwargs)


# Takes address, address:value, address:value ... (satoshis), fee(satoshis)
def preparemultitx(frm, *args, **kwargs):
    tv, fee = args[:-1], int(args[-1])
    outs = []
    outvalue = 0
    for a in tv:
        outs.append(a)
        outvalue += int(a.split(":")[1])

    u = unspent(frm, **kwargs)
    u2 = select(u, int(outvalue)+int(fee))
    argz = u2 + outs + [frm, fee]
    return mksend(*argz, **kwargs)


# BIP32 hierarchical deterministic multisig script
def bip32_hdm_script(*args):
    if len(args) == 3:
        keys, req, path = args
    else:
        i, keys, path = 0, [], []
        while len(args[i]) > 40:
            keys.append(args[i])
            i += 1
        req = int(args[i])
        path = map(int, args[i+1:])
    pubs = sorted(map(lambda x: bip32_descend(x, path), keys))
    return mk_multisig_script(pubs, req)


# BIP32 hierarchical deterministic multisig address
def bip32_hdm_addr(*args):
    return scriptaddr(bip32_hdm_script(*args))


# Setup a coinvault transaction
def setup_coinvault_tx(tx, script):
    txobj = deserialize(tx)
    N = deserialize_script(script)[-2]
    for inp in txobj["ins"]:
        inp["script"] = serialize_script([None] * (N+1) + [script])
    return serialize(txobj)


# Sign a coinvault transaction
def sign_coinvault_tx(tx, priv):
    pub = privtopub(priv)
    txobj = deserialize(tx)
    subscript = deserialize_script(txobj['ins'][0]['script'])
    oscript = deserialize_script(subscript[-1])
    k, pubs = oscript[0], oscript[1:-2]
    for j in range(len(txobj['ins'])):
        scr = deserialize_script(txobj['ins'][j]['script'])
        for i, p in enumerate(pubs):
            if p == pub:
                scr[i+1] = multisign(tx, j, subscript[-1], priv)
        if len(filter(lambda x: x, scr[1:-1])) >= k:
            scr = [None] + filter(lambda x: x, scr[1:-1])[:k] + [scr[-1]]
        txobj['ins'][j]['script'] = serialize_script(scr)
    return serialize(txobj)


# Inspects a transaction
def inspect(tx, **kwargs):
    d = deserialize(tx)
    isum = 0
    ins = {}
    for _in in d['ins']:
        h = _in['outpoint']['hash']
        i = _in['outpoint']['index']
        prevout = deserialize(fetchtx(h, **kwargs))['outs'][i]
        isum += prevout['value']
        a = script_to_address(prevout['script'])
        ins[a] = ins.get(a, 0) + prevout['value']
    outs = []
    osum = 0
    for _out in d['outs']:
        outs.append({'address': script_to_address(_out['script']),
                     'value': _out['value']})
        osum += _out['value']
    return {
        'fee': isum - osum,
        'outs': outs,
        'ins': ins
    }


def merkle_prove(txhash):
    blocknum = str(get_block_height(txhash))
    header = get_block_header_data(blocknum)
    hashes = get_txs_in_block(blocknum)
    i = hashes.index(txhash)
    return mk_merkle_proof(header, hashes, i)



#!/usr/bin/python
from _functools import reduce

import copy

from bitcoin.bech32 import bech32decode, BECH32_PREFIXES, bech32encode

from bitcoin.main import *
### Hex to bin converter and vice versa for objects


def json_is_base(obj, base):
    if not is_python2 and isinstance(obj, bytes):
        return False

    alpha = get_code_string(base)
    if isinstance(obj, string_types):
        for i in range(len(obj)):
            if alpha.find(obj[i]) == -1:
                return False
        return True
    elif isinstance(obj, int_types) or obj is None:
        return True
    elif isinstance(obj, list):
        for i in range(len(obj)):
            if not json_is_base(obj[i], base):
                return False
        return True
    else:
        for x in obj:
            if not json_is_base(obj[x], base):
                return False
        return True


def json_changebase(obj, changer):
    if isinstance(obj, string_or_bytes_types):
        return changer(obj)
    elif isinstance(obj, int_types) or obj is None:
        return obj
    elif isinstance(obj, list):
        return [json_changebase(x, changer) for x in obj]
    return dict((x, json_changebase(obj[x], changer)) for x in obj)

# Transaction serialization and deserialization

def deserialize(tx):
    if is_hexilified(tx):
        #tx = bytes(bytearray.fromhex(tx))
        return json_changebase(deserialize(binascii.unhexlify(tx)),
                              lambda x: safe_hexlify(x))
    # http://stackoverflow.com/questions/4851463/python-closure-write-to-variable-in-parent-scope
    # Python's scoping rules are demented, requiring me to make pos an object
    # so that it is call-by-reference
    pos = [0]

    def read_as_int(bytez):
        pos[0] += bytez
        return decode(tx[pos[0]-bytez:pos[0]][::-1], 256)

    def read_var_int():
        pos[0] += 1

        val = from_byte_to_int(tx[pos[0]-1])
        if val < 253:
            return val
        return read_as_int(pow(2, val - 252))

    def read_bytes(bytez):
        pos[0] += bytez
        return tx[pos[0]-bytez:pos[0]]

    def read_var_string():
        size = read_var_int()
        return read_bytes(size)

    obj = {"ins": [], "outs": []}
    obj["version"] = read_as_int(4)
    ins = read_var_int()

    " begin segwit "
    segwit_flag = False
    if not ins:
        segwit_flag = read_var_int()
        ins = read_var_int()
    " end segwit "

    for i in range(ins):
        obj["ins"].append({
            "outpoint": {
                "hash": read_bytes(32)[::-1],
                "index": read_as_int(4)
            },
            "script": read_var_string(),
            "sequence": read_as_int(4)
        })
    outs = read_var_int()
    for i in range(outs):
        obj["outs"].append({
            "value": read_as_int(8),
            "script": read_var_string()
        })
    if segwit_flag:
        obj["segwit"] = True
        for i in range(ins):
            howmany = read_var_int()
            if howmany:
                obj['ins'][i]['txinwitness'] = [read_var_string() for x in range(0, howmany)]

    obj["locktime"] = read_as_int(4)
    return obj


def serialize(txobj):
    SEGWIT_MARKER = b'\x00'
    SEGWIT_FLAG = b'\x01'
    o = []
    if json_is_base(txobj, 16):
        json_changedbase = json_changebase(txobj, lambda x: binascii.unhexlify(x))
        return str(binascii.hexlify(serialize(json_changedbase)).decode())
    o.append(encode(txobj["version"], 256, 4)[::-1])

    " begin segwit "
    segwit = txobj.get('segwit', False)
    " end segwit "

    o.append(num_to_var_int(len(txobj["ins"])))
    for inp in txobj["ins"]:
        inp['script'] = inp.get('script', '')
        " begin segwit "
        segwit = bool(inp.get("txinwitness") != None) if not segwit else segwit
        " end segwit "
        o.append(inp["outpoint"]["hash"][::-1])
        o.append(encode(inp["outpoint"]["index"], 256, 4)[::-1])
        out_len = num_to_var_int(len(inp["script"]))
        script = inp['script'] if inp.get('script') else b''
        o.append(out_len + script)
        o.append(encode(inp.get('sequence', 4294967295), 256, 4)[::-1])
    o.append(num_to_var_int(len(txobj["outs"])))
    for out in txobj["outs"]:
        out['script'] = out['script']
        o.append(encode(out["value"], 256, 8)[::-1])
        o.append(num_to_var_int(len(out["script"]))+out["script"])

    " begin segwit "
    if segwit:
        o.insert(1, SEGWIT_MARKER + SEGWIT_FLAG)
        for inp in txobj["ins"]:
            if not isinstance(inp.get('txinwitness', None), list):
                o.append(num_to_var_int(0))
            else:
                if len(inp.get('txinwitness')):
                    o.append(num_to_var_int(len(inp.get('txinwitness'))))
                    for i in inp.get('txinwitness'):
                        o.append(num_to_var_int(len(i)))
                        if i:
                            o.append(i)
                else:
                    o.append(num_to_var_int(2))
                    o.append(num_to_var_int(0) * 2)

    " end segwit "

    o.append(encode(txobj["locktime"], 256, 4)[::-1])
    return b''.join(o)


def signature_form(tx, i, script, hashcode=SIGHASH_ALL):
    i, hashcode = int(i), int(hashcode)
    if isinstance(tx, string_or_bytes_types):
        sform = signature_form(deserialize(tx), i, script, hashcode)
        return serialize(sform)

    newtx = copy.deepcopy(tx)
    for inp in newtx["ins"]:
        inp["script"] = ""
    newtx["ins"][i]["script"] = script

    if hashcode == SIGHASH_SINGLE:
        newtx["outs"] = newtx["outs"][:len(newtx["ins"])]
        for out in newtx["outs"][:len(newtx["ins"]) - 1]:
            out['value'] = 2**64 - 1
            out['script'] = ""
    elif hashcode & SIGHASH_NONE:
        newtx["outs"] = []
    if hashcode & SIGHASH_ANYONECANPAY:
        newtx["ins"] = [newtx["ins"][i]]
    return newtx


# Making the actual signatures


def encode_num(n):
    h = binascii.hexlify(encode(n,256))
    b = binascii.unhexlify(h)
    if ord(b[0]) < 0x80:
        return h
    else:
        return '00' + h

def der_encode_sig(v, r, s):
    b1, b2 = safe_hexlify(encode(r, 256)), safe_hexlify(encode(s, 256))
    if len(b1) and b1[0] in '89abcdef':
        b1 = '00' + b1
    if len(b2) and b2[0] in '89abcdef':
        b2 = '00' + b2
    left = '02'+encode(len(b1)//2, 16, 2)+b1
    right = '02'+encode(len(b2)//2, 16, 2)+b2
    return '30'+encode(len(left+right)//2, 16, 2)+left+right

def der_decode_sig(sig):
    leftlenbytes = decode(sig[6:8], 16)
    leftlen = leftlenbytes*2
    left = sig[8:8+leftlen]
    rightlenbytes = decode(sig[10+leftlen:12+leftlen], 16)
    rightlen = rightlenbytes*2
    right = sig[12+leftlen:12+leftlen+rightlen]
    return (leftlenbytes, decode(left, 16), rightlenbytes, decode(right, 16))

def is_bip66(sig):
    """Checks hex DER sig for BIP66 consistency"""
    #https://raw.githubusercontent.com/bitcoin/bips/master/bip-0066.mediawiki
    #0x30  [total-len]  0x02  [R-len]  [R]  0x02  [S-len]  [S]  [sighash]
    sig = bytearray.fromhex(sig) if re.match('^[0-9a-fA-F]*$', sig) else bytearray(sig)
    if (sig[0] == 0x30) and (sig[1] == len(sig)-2):     # check if sighash is missing
            sig.extend(b"\1")		                   	# add SIGHASH_ALL for testing
    #assert (sig[-1] & 124 == 0) and (not not sig[-1]), "Bad SIGHASH value"

    if len(sig) < 9 or len(sig) > 73: return False
    if (sig[0] != 0x30): return False
    if (sig[1] != len(sig)-3): return False
    rlen = sig[3]
    if (5+rlen >= len(sig)): return False
    slen = sig[5+rlen]
    if (rlen + slen + 7 != len(sig)): return False
    if (sig[2] != 0x02): return False
    if (rlen == 0): return False
    if (sig[4] & 0x80): return False
    if (rlen > 1 and (sig[4] == 0x00) and not (sig[5] & 0x80)): return False
    if (sig[4+rlen] != 0x02): return False
    if (slen == 0): return False
    if (sig[rlen+6] & 0x80): return False
    if (slen > 1 and (sig[6+rlen] == 0x00) and not (sig[7+rlen] & 0x80)):
        return False
    return True

def txhash(tx, hashcode=None):
    if isinstance(tx, str) and re.match('^[0-9a-fA-F]*$', tx):
        tx = changebase(tx, 16, 256)
    if hashcode is not None:
        return dbl_sha256(from_string_to_bytes(tx) + encode(int(hashcode), 256, 4)[::-1])
    else:
        return safe_hexlify(bin_dbl_sha256(tx)[::-1])


def bin_txhash(tx, hashcode=None):
    return binascii.unhexlify(txhash(tx, hashcode))


def ecdsa_tx_sign(tx, priv, hashcode=SIGHASH_ALL):
    rawsig = ecdsa_raw_sign(bin_txhash(tx, hashcode), priv)
    return der_encode_sig(*rawsig)+encode(hashcode, 16, 2)


def ecdsa_tx_verify(tx, sig, pub, hashcode=SIGHASH_ALL):
    return ecdsa_raw_verify(bin_txhash(tx, hashcode), der_decode_sig(sig), pub)


def ecdsa_tx_recover(tx, sig, hashcode=SIGHASH_ALL):
    z = bin_txhash(tx, hashcode)
    rlen, r, slen, s = der_decode_sig(sig)
    left = ecdsa_raw_recover(z, (0, r, s))
    right = ecdsa_raw_recover(z, (1, r, s))
    return (encode_pubkey(left, 'hex'), encode_pubkey(right, 'hex'))

# Scripts


def mk_pubkey_script(addr):
    # Keep the auxiliary functions around for altcoins' sake
    return '76a914' + b58check_to_hex(addr) + '88ac'


def mk_scripthash_script(addr):
    return 'a914' + b58check_to_hex(addr) + '87'

# Address representation to output script


def address_to_script(addr):
    if addr[0] in ('2', '3', 'S', 'r'):
        return mk_scripthash_script(addr)
    else:
        return mk_pubkey_script(addr)

# Output script to address representation


def scripthash_to_address(script: str, b58_vbyte: int, scripthash_byte: int, b32_prefix: str):
    if re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    if script[:2] == b'\x00\x20':
        return bech32encode(script.hex(), b32_prefix)
    if script[:2] == b'\x00\x14':
        return bech32encode(script.hex(), b32_prefix)

    if script[:3] == b'\x76\xa9\x14' and script[-2:] == b'\x88\xac' and len(script) == 25:
        return bin_to_b58check(script[3:-2], b58_vbyte)  # pubkey hash addresses
    else:
        return bin_to_b58check(script[2:-1], scripthash_byte)


def script_to_address(script, vbyte=0):
    if re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    if script[:3] == b'\x76\xa9\x14' and script[-2:] == b'\x88\xac' and len(script) == 25:
        return bin_to_b58check(script[3:-2], vbyte)  # pubkey hash addresses
    else:
        if vbyte in [111, 196]:
            # Testnet
            scripthash_byte = 196
        elif vbyte == 0:
            # Mainnet
            scripthash_byte = 5
        else:
            scripthash_byte = vbyte
        # BIP0016 scripthash addresses
        return bin_to_b58check(script[2:-1], scripthash_byte)


def p2sh_scriptaddr(script, magicbyte=5):
    if re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    return hex_to_b58check(hash160(script), magicbyte)
scriptaddr = p2sh_scriptaddr


def deserialize_script(script):
    if isinstance(script, str) and re.match('^[0-9a-fA-F]*$', script):
       return json_changebase(deserialize_script(binascii.unhexlify(script)),
                              lambda x: safe_hexlify(x))
    out, pos = [], 0
    while pos < len(script):
        code = from_byte_to_int(script[pos])
        if code == 0:
            out.append(None)
            pos += 1
        elif code <= 75:
            out.append(script[pos+1:pos+1+code])
            pos += 1 + code
        elif code <= 78:
            szsz = pow(2, code - 76)
            sz = decode(script[pos+szsz: pos:-1], 256)
            out.append(script[pos + 1 + szsz:pos + 1 + szsz + sz])
            pos += 1 + szsz + sz
        elif code <= 96:
            out.append(code - 80)
            pos += 1
        else:
            out.append(code)
            pos += 1
    return out


def serialize_script_unit(unit):
    if isinstance(unit, int):
        if unit < 16:
            return from_int_to_byte(unit + 80)
        else:
            return from_int_to_byte(unit)
    elif unit is None:
        return b'\x00'
    else:
        if len(unit) <= 75:
            return from_int_to_byte(len(unit))+unit
        elif len(unit) < 256:
            return from_int_to_byte(76)+from_int_to_byte(len(unit))+unit
        elif len(unit) < 65536:
            return from_int_to_byte(77)+encode(len(unit), 256, 2)[::-1]+unit
        else:
            return from_int_to_byte(78)+encode(len(unit), 256, 4)[::-1]+unit


if is_python2:
    def serialize_script(script):
        if json_is_base(script, 16):
            return binascii.hexlify(serialize_script(json_changebase(script,
                                    lambda x: binascii.unhexlify(x))))
        return ''.join(map(serialize_script_unit, script))
else:
    def serialize_script(script):
        if json_is_base(script, 16):
            return safe_hexlify(serialize_script(json_changebase(script,
                                    lambda x: binascii.unhexlify(x))))

        result = bytes()
        for b in map(serialize_script_unit, script):
            result += b if isinstance(b, bytes) else bytes(b, 'utf-8')
        return result


def mk_OPCS_multisig_script(form):
    script = []
    OP_CODESEPARATOR = from_int_to_byte(0xAB)
    OP_CHECKSIG = from_int_to_byte(0xAC)
    OP_CHECKSIGVERIFY = from_int_to_byte(0xAD)
    OP_CHECKMULTISIG = from_int_to_byte(0xAE)
    OP_CHECKMULTISIGVERIFY = from_int_to_byte(0xAF)
    """
    uses OP_CODESEPARATOR to build the multisig redeem script
    form:

    {
    'keys': ['hex_pubkey_1',
             'hex_pubkey_2',
             'hex_pubkey_3'
    'schema': [
            {
                'reqs': 1,
                'keys': [0],
            },
            {
                'reqs': 1,
                'keys': [1, 2],
            }
        ]
    }

    """
    for i, s in enumerate(form['schema']):
        subscript = []
        for k in s['keys']:
            subscript.append(from_int_to_byte(len(form['keys'][k]) // 2))
            subscript.append(safe_from_hex(form['keys'][k]))
        if len(s['keys']) == 1:
            subscript.append(OP_CHECKSIG if i == len(s) -1 else OP_CHECKSIGVERIFY)
        elif len(s['keys']) > 1:
            subscript.insert(from_int_to_byte(s['reqs']), 0)
            subscript.insert(from_int_to_byte(len(s['keys'])), len(s['keys']) + 1)
            subscript.append(OP_CHECKMULTISIG if i == len(s) -1 else OP_CHECKMULTISIGVERIFY)
        else:
            raise ValueError('no keys condition')
        if i != len(s) - 1:
            subscript.append(OP_CODESEPARATOR)
        script = script + subscript
    res = binascii.hexlify(b''.join(script)).decode('ascii')
    return res


def mk_multisig_script(*args):  # [pubs],k or pub1,pub2...pub[n],k
    if isinstance(args[0], list):
        pubs, k = args[0], int(args[1])
    elif len(args) == 1 and isinstance(args[0], dict):
        return mk_OPCS_multisig_script(args[0])
    else:
        pubs = list(filter(lambda x: len(str(x)) >= 32, args))
        k = int(args[len(pubs)])
    return serialize_script([k]+pubs+[len(pubs)]+[0xae])

# Signing and verifying


def verify_tx_input(tx, i, script, sig, pub):
    if re.match('^[0-9a-fA-F]*$', tx):
        tx = binascii.unhexlify(tx)
    if re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    if not re.match('^[0-9a-fA-F]*$', sig):
        sig = safe_hexlify(sig)
    hashcode = decode(sig[-2:], 16)
    modtx = signature_form(tx, int(i), script, hashcode)
    return ecdsa_tx_verify(modtx, sig, pub, hashcode)


def sign(tx, i, priv, hashcode=SIGHASH_ALL):
    i = int(i)
    txobj = tx if isinstance(tx, dict) else deserialize(tx)
    if not isinstance(tx, dict) and ((not is_python2 and isinstance(re, bytes)) or not re.match('^[0-9a-fA-F]*$', tx)):
        return binascii.unhexlify(sign(safe_hexlify(tx), i, priv))
    if len(priv) <= 33:
        priv = safe_hexlify(priv)
    pub = privkey_to_pubkey(priv)
    address = pubkey_to_address(pub)
    signing_tx = signature_form(tx, i, mk_pubkey_script(address), hashcode)
    sig = ecdsa_tx_sign(signing_tx, priv, hashcode)
    txobj["ins"][i]['script'] = serialize_script([sig, pub])
    return serialize(txobj)

def p2pk_sign(tx, i, priv, hashcode=SIGHASH_ALL):
    i = int(i)
    txobj = tx if isinstance(tx, dict) else deserialize(tx)
    if not isinstance(tx, dict) and ((not is_python2 and isinstance(re, bytes)) or not re.match('^[0-9a-fA-F]*$', tx)):
        return binascii.unhexlify(sign(safe_hexlify(tx), i, priv))
    if len(priv) <= 33:
        priv = safe_hexlify(priv)
    pub = privkey_to_pubkey(priv)
    signing_tx = signature_form(tx, i, '21' + pub + 'ac', hashcode)
    sig = ecdsa_tx_sign(signing_tx, priv, hashcode)
    txobj["ins"][i]['script'] = serialize_script([sig])
    return serialize(txobj)


def signall(tx, priv):
    # if priv is a dictionary, assume format is
    # { 'txinhash:txinidx' : privkey }
    if isinstance(priv, dict):
        for e, i in enumerate(deserialize(tx)["ins"]):
            k = priv["%s:%d" % (i["outpoint"]["hash"], i["outpoint"]["index"])]
            tx = sign(tx, e, k)
    else:
        for i in range(len(deserialize(tx)["ins"])):
            tx = sign(tx, i, priv)
    return tx


def multisign(tx, i, script, pk, hashcode=SIGHASH_ALL):
    if re.match('^[0-9a-fA-F]*$', tx):
        tx = binascii.unhexlify(tx)
    if re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    modtx = signature_form(tx, i, script, hashcode)
    return ecdsa_tx_sign(modtx, pk, hashcode)


def is_inp(arg):
    return len(arg) > 64 or "output" in arg or "outpoint" in arg


def mktx(*args, **kwargs):
    # [in0, in1...],[out0, out1...] or in0, in1 ... out0 out1 ...
    ser = kwargs.get('serialize', True)
    ins, outs = [], []
    for arg in args:
        if isinstance(arg, list):
            for a in arg: (ins if is_inp(a) else outs).append(a)
        else:
            (ins if is_inp(arg) else outs).append(arg)

    txobj = {"locktime": kwargs.get('locktime', 0), "version": 1, "ins": [], "outs": []}
    for i in ins:

        " begin segwit "
        seg_input = isinstance(i, dict) and i.get('segregated')
        sequence = isinstance(i, dict) and i.get('sequence', None)
        " end segwit "

        if isinstance(i, dict) and "outpoint" in i:
            txobj["ins"].append(i)
        else:
            if isinstance(i, dict) and "output" in i:
                i = i["output"]
            txobj["ins"].append({
                "outpoint": {"hash": i[:64], "index": int(i[65:])},
                "script": "",
                "sequence": 4294967295 if not sequence and sequence != 0 else sequence
            })
        if seg_input:
            txobj["ins"][-1].update({'txinwitness': []})

    for o in outs:
        if isinstance(o, string_or_bytes_types):
            addr = o[:o.find(':')]
            val = int(o[o.find(':')+1:])
            o = {}
            if re.match('^[0-9a-fA-F]*$', addr):
                o["script"] = addr
            else:
                o["address"] = addr
            o["value"] = val
        outobj = {}
        if "address" in o:
            if any(map(lambda x: o["address"].lower().startswith(x), BECH32_PREFIXES)):
                outobj["script"] = bech32decode(o["address"].lower())
            else:
                outobj["script"] = address_to_script(o["address"])
        elif "script" in o:
            outobj["script"] = o["script"]
        else:
            raise Exception("Could not find 'address' or 'script' in output.")
        outobj["value"] = o["value"]
        txobj["outs"].append(outobj)
    return txobj if not ser else serialize(txobj)


def select(unspent, value):
    value = int(value)
    high = [u for u in unspent if u["value"] >= value]
    high.sort(key=lambda u: u["value"])
    low = [u for u in unspent if u["value"] < value]
    low.sort(key=lambda u: -u["value"])
    if len(high):
        return [high[0]]
    i, tv = 0, 0
    while tv < value and i < len(low):
        tv += low[i]["value"]
        i += 1
    if tv < value:
        raise Exception("Not enough funds")
    return low[:i]

# Only takes inputs of the form { "output": blah, "value": foo }


def mksend(*args, **kwargs):
    argz, change, fee = args[:-2], args[-2], int(args[-1])
    ins, outs = [], []
    for arg in argz:
        if isinstance(arg, list):
            for a in arg:
                (ins if is_inp(a) else outs).append(a)
        else:
            (ins if is_inp(arg) else outs).append(arg)

    isum = sum([i["value"] for i in ins])
    osum, outputs2 = 0, []
    for o in outs:
        if isinstance(o, string_types):
            o2 = {
                "address": o[:o.find(':')],
                "value": int(o[o.find(':')+1:])
            }
        else:
            o2 = o
        outputs2.append(o2)
        osum += o2["value"]

    if isum < osum+fee:
        raise Exception("Not enough money")
    elif isum > osum+fee+5430:
        outputs2 += [{"address": change, "value": isum-osum-fee}]

    return mktx(ins, outputs2, **kwargs)

def mk_opreturn(msg, rawtx=None, json=0):
    def op_push(data):
        import struct
        if len(data) < 0x4c:
            return from_int_to_byte(len(data)) + data
        elif len(data) < 0xff:
            return from_int_to_byte(76) + struct.pack('<B', len(data)) + from_string_to_bytes(data)
        elif len(data) < 0xffff:
            return from_int_to_byte(77) + struct.pack('<H', len(data)) + from_string_to_bytes(data)
        elif len(data) < 0xffffffff:
            return from_int_to_byte(78) + struct.pack('<I', len(data)) + from_string_to_bytes(data)
        else:
            raise Exception("Input data error. Rawtx must be hex chars" \
                            + "0xffffffff > len(data) > 0 ??")

    orhex = safe_hexlify(b'\x6a' + op_push(msg))
    orjson = {'script' : orhex, 'value' : 0}
    if rawtx is not None:
        try:
            txo = deserialize(rawtx)
            if not 'outs' in txo.keys(): raise Exception("OP_Return cannot be the sole output!")
            txo['outs'].append(orjson)
            newrawtx = serialize(txo)
            return newrawtx
        except:
            raise Exception("Raw Tx Error!")
    return orhex if not json else orjson


import bitcoin as bc
import sys
import unittest

class TestStealth(unittest.TestCase):

    def setUp(self):
        
        if sys.getrecursionlimit() < 1000:
            sys.setrecursionlimit(1000)
        
        self.addr = 'vJmtjxSDxNPXL4RNapp9ARdqKz3uJyf1EDGjr1Fgqs9c8mYsVH82h8wvnA4i5rtJ57mr3kor1EVJrd4e5upACJd588xe52yXtzumxj'
        self.scan_pub = '025e58a31122b38c86abc119b9379fe247410aee87a533f9c07b189aef6c3c1f52'
        self.scan_priv = '3e49e7257cb31db997edb1cf8299af0f37e2663e2260e4b8033e49d39a6d02f2'
        self.spend_pub = '03616562c98e7d7b74be409a787cec3a912122f3fb331a9bee9b0b73ce7b9f50af'
        self.spend_priv = 'aa3db0cfb3edc94de4d10f873f8190843f2a17484f6021a95a7742302c744748'
        self.ephem_pub = '03403d306ec35238384c7e340393335f9bc9bb4a2e574eb4e419452c4ea19f14b0'
        self.ephem_priv = '9e63abaf8dcd5ea3919e6de0b6c544e00bf51bf92496113a01d6e369944dc091'
        self.shared_secret = 'a4047ee231f4121e3a99a3a3378542e34a384b865a9917789920e1f13ffd91c6'
        self.pay_pub = '02726112ad39cb6bf848b1b1ef30b88e35286bf99f746c2be575f96c0e02a9357c'
        self.pay_priv = '4e422fb1e5e1db6c1f6ab32a7706d368ceb385e7fab098e633c5c5949c3b97cd'
        
        self.testnet_addr = 'waPUuLLykSnY3itzf1AyrQZm42F7KyB7SR5zpfqmnzPXWhx9kXLzV3EcyqzDdpTwngiyCCMUqztS9S1d7XJs3JMt3MsHPDpBCudvx9'
        
    def test_address_encoding(self):

        sc_pub, sp_pub = bc.basic_stealth_address_to_pubkeys(self.addr)
        self.assertEqual(sc_pub, self.scan_pub)
        self.assertEqual(sp_pub, self.spend_pub)
        
        stealth_addr2 = bc.pubkeys_to_basic_stealth_address(sc_pub, sp_pub)
        self.assertEqual(stealth_addr2, self.addr)
        
        magic_byte_testnet = 43
        sc_pub, sp_pub = bc.basic_stealth_address_to_pubkeys(self.testnet_addr)
        self.assertEqual(sc_pub, self.scan_pub)
        self.assertEqual(sp_pub, self.spend_pub)
        
        stealth_addr2 = bc.pubkeys_to_basic_stealth_address(sc_pub, sp_pub, magic_byte_testnet)
        self.assertEqual(stealth_addr2, self.testnet_addr)
        
    def test_shared_secret(self):

        sh_sec = bc.shared_secret_sender(self.scan_pub, self.ephem_priv)
        self.assertEqual(sh_sec, self.shared_secret)

        sh_sec2 = bc.shared_secret_receiver(self.ephem_pub, self.scan_priv)
        self.assertEqual(sh_sec2, self.shared_secret)

    def test_uncover_pay_keys(self):

        pub = bc.uncover_pay_pubkey_sender(self.scan_pub, self.spend_pub, self.ephem_priv)
        pub2 = bc.uncover_pay_pubkey_receiver(self.scan_priv, self.spend_pub, self.ephem_pub)
        self.assertEqual(pub, self.pay_pub)
        self.assertEqual(pub2, self.pay_pub)

        priv = bc.uncover_pay_privkey(self.scan_priv, self.spend_priv, self.ephem_pub)
        self.assertEqual(priv, self.pay_priv)

    def test_stealth_metadata_script(self):

        nonce = int('deadbeef', 16)
        script = bc.mk_stealth_metadata_script(self.ephem_pub, nonce)
        self.assertEqual(script[6:], 'deadbeef' + self.ephem_pub)
        
        eph_pub = bc.ephem_pubkey_from_tx_script(script)
        self.assertEqual(eph_pub, self.ephem_pub)

    def test_stealth_tx_outputs(self):

        nonce = int('deadbeef', 16)
        value = 10**8
        outputs = bc.mk_stealth_tx_outputs(self.addr, value, self.ephem_priv, nonce)

        self.assertEqual(outputs[0]['value'], 0)
        self.assertEqual(outputs[0]['script'], '6a2606deadbeef' + self.ephem_pub)
        self.assertEqual(outputs[1]['address'], bc.pubkey_to_address(self.pay_pub))
        self.assertEqual(outputs[1]['value'], value)
        
        outputs = bc.mk_stealth_tx_outputs(self.testnet_addr, value, self.ephem_priv, nonce, 'testnet')
        
        self.assertEqual(outputs[0]['value'], 0)
        self.assertEqual(outputs[0]['script'], '6a2606deadbeef' + self.ephem_pub)
        self.assertEqual(outputs[1]['address'], bc.pubkey_to_address(self.pay_pub, 111))
        self.assertEqual(outputs[1]['value'], value)

        self.assertRaises(Exception, bc.mk_stealth_tx_outputs, self.testnet_addr, value, self.ephem_priv, nonce, 'btc')
        
        self.assertRaises(Exception, bc.mk_stealth_tx_outputs, self.addr, value, self.ephem_priv, nonce, 'testnet')
 
if __name__ == '__main__':
    unittest.main()



import binascii
import hashlib
import re
import struct
from bitcoin import changebase, is_python2, privkey_to_pubkey, pubkey_to_address, ecdsa_raw_sign, encode, \
    SIGHASH_ALL, SIGHASH_ANYONECANPAY, SIGHASH_SINGLE, SIGHASH_NONE, \
    deserialize, txhash, serialize, sign, mk_pubkey_script, der_encode_sig, serialize_script, \
    deserialize_script, decode, ecdsa_raw_verify, der_decode_sig


def get_hashcode_strategy(hashcode):
    return _STRATEGIES[hashcode]


class HashcodeStrategy():
    def get_inputs_for_sequences(self, tx, i):
        raise NotImplementedError()

    def get_outputs(self, tx, i):
        raise NotImplementedError()

    def get_sequences(self, tx, i):
        raise NotImplementedError()


class SIGHASH_ALL_Strategy(HashcodeStrategy):
    def get_inputs_for_sequences(self, tx, i):
        return tx['ins']

    def get_outputs(self, tx, i):
        return tx['outs']

    def get_sequences(self, tx, i):
        return [struct.pack('<I', (inps['sequence'])) for inps in self.get_inputs_for_sequences(tx, i)]


class SIGHASH_SINGLE_Strategy(HashcodeStrategy):
    def get_inputs_for_sequences(self, tx, i):
        return tx['ins']

    def get_outputs(self, tx, i):
        if len(tx['outs']) > i:
            return [tx['outs'][i]]
        return []

    def get_sequences(self, tx, i):
        return []

class SIGHASH_NONE_Strategy(HashcodeStrategy):
    def get_inputs_for_sequences(self, tx, i):
        return tx['ins']

    def get_outputs(self, tx, i):
        return []

    def get_sequences(self, tx, i):
        return []


class ANYONECANPAY_STRATEGY(HashcodeStrategy):
    def __init__(self, default_strategy):
        self.default = default_strategy

    def get_inputs_for_sequences(self, tx, i):
        return []

    def get_outputs(self, tx, i):
        return self.default.get_outputs(tx, i)

    def get_sequences(self, tx, i):
        return []

_STRATEGIES = {
    SIGHASH_ALL: SIGHASH_ALL_Strategy(),
    SIGHASH_SINGLE: SIGHASH_SINGLE_Strategy(),
    SIGHASH_NONE: SIGHASH_NONE_Strategy(),
    SIGHASH_ALL|SIGHASH_ANYONECANPAY: ANYONECANPAY_STRATEGY(SIGHASH_ALL_Strategy()),
    SIGHASH_SINGLE|SIGHASH_ANYONECANPAY: ANYONECANPAY_STRATEGY(SIGHASH_SINGLE_Strategy()),
    SIGHASH_NONE|SIGHASH_ANYONECANPAY: ANYONECANPAY_STRATEGY(SIGHASH_NONE_Strategy())
}


def is_segwit(tx, hashcode=None):
    if isinstance(tx, str) and re.match('^[0-9a-fA-F]*$', tx):
        tx = changebase(tx, 16, 256)
    return tx[4:6] == b'\x00\x01'


def segwit_signature_form_transaction_dict(tx, i, script, amount, hashcode=SIGHASH_ALL):
    def parse_vout(o):
        return b''.join([struct.pack('<Q', o['value']),
                         struct.pack('B', len(o['script']) // 2),
                         binascii.unhexlify(o['script'])])
    def parse_vin(inp):
        return b''.join([binascii.unhexlify(inp['outpoint']['hash'])[::-1],
                         struct.pack('<I', (inp['outpoint']['index']))])

    vin_outpoint = [binascii.unhexlify(tx['ins'][i]['outpoint']['hash'])[::-1],
                    struct.pack('<I', (tx['ins'][i]['outpoint']['index']))]
    hashcode_strategy = get_hashcode_strategy(hashcode)
    outpoints = [parse_vin(inp) for inp in hashcode_strategy.get_inputs_for_sequences(tx, i)]
    sequences = hashcode_strategy.get_sequences(tx, i)
    outputs_to_sign = hashcode_strategy.get_outputs(tx, i)
    outputs = [parse_vout(out) for out in outputs_to_sign]

    hash_outputs = hashlib.sha256(hashlib.sha256(b''.join(outputs)).digest()).digest() if outputs_to_sign else b'\x00'*32
    hash_sequences = hashlib.sha256(hashlib.sha256(b''.join(sequences)).digest()).digest() if sequences else b'\x00'*32
    hash_outpoints = hashlib.sha256(hashlib.sha256(b''.join(outpoints)).digest()).digest() if outpoints else b'\x00'*32

    preimage = [struct.pack('<I', tx['version']),
                hash_outpoints,
                hash_sequences,
                b''.join(vin_outpoint),
                struct.pack('B', len(script) // 2),
                binascii.unhexlify(script),
                struct.pack('<Q', amount),
                struct.pack('<I', tx['ins'][i]['sequence']),
                hash_outputs,
                struct.pack('<I', tx['locktime']),
                struct.pack('<I', hashcode)]
    return binascii.hexlify(b''.join(preimage)).decode('ascii')


def segwit_signature_form(tx, i, script, amount, hashcode=SIGHASH_ALL):
    return segwit_signature_form_transaction_dict(
        deserialize(tx), i, script, amount, hashcode=hashcode
    )


def segwit_txhash(tx):
    if isinstance(tx, str) and re.match('^[0-9a-fA-F]*$', tx):
        tx = changebase(tx, 16, 256)
    tx_hash = txhash(tx)
    tx = strip_witness_data(tx)
    tx_id = txhash(tx)
    return {'hash': tx_hash, 'txid': tx_id}


def strip_witness_data(tx):
    return serialize(
        strip_witness_deserialized_data(
            deserialize(tx)
        )
    )


def strip_witness_deserialized_data(tx_dict):
    tx_dict.pop('segwit', '')
    for inp in tx_dict['ins']:
        inp.pop('txinwitness', '')
    return tx_dict


def segwit_sign(tx, i, priv, amount, hashcode=SIGHASH_ALL, script=None, separator_index=None):
    i = int(i)
    txobj = tx if isinstance(tx, dict) else deserialize(tx)
    if not isinstance(tx, dict) and ((not is_python2 and isinstance(re, bytes)) or not re.match('^[0-9a-fA-F]*$', tx)):
        return binascii.unhexlify(sign(binascii.hexlify(tx), i, priv))
    if len(priv) <= 33:
        priv = binascii.hexlify(priv)
    pub = privkey_to_pubkey(priv)
    address = pubkey_to_address(pub)
    wscript = mk_pubkey_script(address) if not script else script
    stripped_script = segwit_strip_script_separator(wscript, separator_index)
    signing_tx = segwit_signature_form(tx, i, stripped_script, amount, hashcode=hashcode)
    rawsig = ecdsa_raw_sign(hashlib.sha256(hashlib.sha256(binascii.unhexlify(signing_tx)).digest()).hexdigest(), priv)
    sig = der_encode_sig(*rawsig)+encode(hashcode, 16, 2)
    txobj['ins'][i]['txinwitness'] = [sig, pub if not script else script]
    return serialize(txobj)


def apply_multisignatures(*args):
    # tx,i,script,sigs OR tx,i,script,sig1,sig2...,sig[n]
    tx, i, script = args[0], int(args[1]), args[2]
    sigs = args[3] if isinstance(args[3], list) else list(args[3:])

    if isinstance(script, str) and re.match('^[0-9a-fA-F]*$', script):
        script = binascii.unhexlify(script)
    sigs = [binascii.unhexlify(x) if x[:2] == '30' else x for x in sigs]
    if isinstance(tx, str) and re.match('^[0-9a-fA-F]*$', tx):
        signed_tx = apply_multisignatures(binascii.unhexlify(tx), i, script, sigs)
        return binascii.hexlify(signed_tx)

    # Not pushing empty elements on the top of the stack if passing no
    # script (in case of bare multisig inputs there is no script)
    script_blob = [] if script.__len__() == 0 else [script]

    txobj = deserialize(tx)
    txobj["ins"][i]["script"] = serialize_script([None]+sigs+script_blob)
    return serialize(txobj)


def apply_segwit_multisignatures(tx, i, witness_program, signatures, dummy=True, nested=False):
    txobj = deserialize(tx)
    signed = apply_segwit_multisignatures_deserialized_data(
        txobj, i, witness_program, signatures, dummy=dummy, nested=nested
    )
    return serialize(signed)


def apply_segwit_multisignatures_deserialized_data(
    deserialized_tx, i, witness_program, signatures, dummy=True, nested=True
):
    o = [""] + signatures + [witness_program] if dummy else signatures + [witness_program]
    deserialized_tx['ins'][i]['txinwitness'] = o
    if nested:
        redeem_script = hashlib.sha256(binascii.unhexlify(
            witness_program
        )).hexdigest()
        length = len(redeem_script) // 2
        redeem_script = serialize_script(
            [length + 2, None, redeem_script])
        deserialized_tx["ins"][i]["script"] = redeem_script
    return deserialized_tx


def segwit_strip_script_separator(script, index=0):
    if index == None:
        return script
    OP_CODESEPARATOR = 171
    def get_pos(script, index):
        i = 0
        for x in range(0, index):
            try:
                i = script.index(OP_CODESEPARATOR, i)
            except ValueError:
                return i
        return i + 1
    deserialized_script = deserialize_script(str(script))
    pos = get_pos(deserialized_script, index)
    return serialize_script(deserialized_script[pos:])


def segwit_multisign(tx, i, script, pk, amount, hashcode=SIGHASH_ALL, separator_index=None):
    wscript = segwit_strip_script_separator(script, index=separator_index)
    signing_tx = segwit_signature_form(tx, i, wscript, amount, hashcode=hashcode)
    rawsig = ecdsa_raw_sign(hashlib.sha256(hashlib.sha256(
        binascii.unhexlify(signing_tx)).digest()).hexdigest(), pk)
    sig = der_encode_sig(*rawsig)+encode(hashcode, 16, 2)
    return sig


def segwit_verify_tx_input(tx, i, script, sig, pub, amount):
    hashcode = decode(sig[-2:], 16)
    signing_tx = segwit_signature_form(tx, int(i), script, amount, hashcode=hashcode)
    hashed_signing_tx = hashlib.sha256(hashlib.sha256(binascii.unhexlify(signing_tx)).digest()).hexdigest()
    return ecdsa_raw_verify(hashed_signing_tx, der_decode_sig(sig), pub)



from unittest import TestCase

from bitcoin import deserialize, mktx, BECH32_BITCOIN_TESTNET_PREFIX, BECH32_BITCOIN_REGTEST_PREFIX, \
    BECH32_BITCOIN_PREFIX, BECH32_CUSTOM_SIGNET_TESTNET_PREFIX
from bitcoin.bech32 import bech32encode, bech32decode, BECH32_CUSTOM_SIGNET_PREFIX


class TestBech32(TestCase):
    def setUp(self):
        pass

    def test_encode_decode_addresses(self):
        addrs = ['bcrt1qs4xe0x388ygy7drku9gkf2n0emcwj5xms0tnkw']
        print('Decoding')
        for addr in addrs:
            decoded = bech32decode(addr)
            print(decoded)
            assert addr == bech32encode(decoded, prefix=BECH32_BITCOIN_REGTEST_PREFIX)

    def test_encode_decode_mainnet_addresses(self):
        addrs = [
            'bc1qwqdg6squsna38e46795at95yu9atm8azzmyvckulcc7kytlcckxswvvzej',
            'bc1qepn5e55urrelwar3jnzz6w853lxz4vfpysskmmat9mk36u46cvvqtnala0',
            'bc1qurj278havq0phe7y3heufmck246xrnxhtyx5sa92pytky655pr6qsqzzxa',
            'bc1q9909xggg842c58v3cncp9l35vv65gxx8mu3483x3w8xvk8a8ng0stfyrd9',
            'bc1qakft6s8shk0dgunet53pzapsuxc7ewagfj77l6j8358xt6sumtgq9g650j',
            'bc1qkmtl43j4kqucux3tgr52wx9jqfqwy0xjyn89rldu8pm2ksneynksul2ta2',
            'bc1qfa4ycsdv6uv774zjrtvxrp55qvg439ykre3ttm56vw2dhvk8hhnssjjyzs'
        ]
        print('Decoding')
        for addr in addrs:
            decoded = bech32decode(addr)
            assert addr == bech32encode(decoded, prefix=BECH32_BITCOIN_PREFIX)

    def test_encode_decode_custom_signet_addresses(self):
        addrs = [
            'cbc1q782kzgu4s4l2mke75rta8u0uakyh5qymxduxxt',
            'cbc1qec0ye5g7ldmmfs88597j7tdl63c9xekrfszwrg',
            'cbc1qcsupznu7kta43fmrsu4dkcwargwwudd05rljhg',
            'cbc1q3f9fder9ss9x25l8auu0pm8gaa25p7u7m5yzlk',
            'cbc1qclg2sltl80ux22rxsx3mqmh2wpjd46gltvenva',
            'cbc1qs2fmhkg3cx9fxvy8gtq8ppxq0qlcccck4ghlz2',
        ]
        print('Decoding')
        for addr in addrs:
            decoded = bech32decode(addr)
            assert addr == bech32encode(decoded, prefix=BECH32_CUSTOM_SIGNET_PREFIX)

    def test_encode_decode(self):
        transaction = "010000000001019fa7fad392a29f74d124a8770d24e23f0fcc3e3007e756e4114e4b950a5cf37a0200000000f" \
                      "fffffff02402d0e20000000001976a914c0123fd352e7f53a9c4c0e512c490e1a6817dffa88acb19066000000" \
                      "0000220020701a8d401c84fb13e6baf169d59684e17abd9fa216c8cc5b9fc63d622ff8c58d040048304502210" \
                      "0f227d07b7ca8bef3b9d591fc46803b1fc6b41290b3b229064726eee5c3d6fc8b02206450b248d4d035a121e6" \
                      "92d7a0e1b720801e1fd30255f5d1f099dbca61eb92dc01483045022100966181ecf84d9a75be1c7654a29ce2a" \
                      "1cba32b2bfb9988d6564ae3bc7358420702200a5e54430e1a9c4537c84d9159243cbfe020fbac2fa7e09b8cd6" \
                      "efb08c984432016952210375e00eb72e29da82b89367947f29ef34afb75e8654f6ea368e0acdfd92976b7c210" \
                      "3a1b26313f430c4b15bb1fdce663207659d8cac749a0e53d70eff01874496feff2103c96d495bfdd5ba4145e3" \
                      "e046fee45e84a8a48ad05bd8dbb395c011a32cf9f88053ae00000000"

        expected_address = "bc1qwqdg6squsna38e46795at95yu9atm8azzmyvckulcc7kytlcckxswvvzej"
        decoded_expected_address = bech32decode(expected_address)
        deserialized = deserialize(transaction)
        assert decoded_expected_address == deserialized['outs'][1]['script']
        re_encoded_address = bech32encode(deserialized['outs'][1]['script'])
        assert re_encoded_address == expected_address

    def test_testnet_encode_decode(self):
        transaction = "01000000000104acb16c8695c04bfff4857d7b6115ca1a539775c4ea396804038c0cd1acbd595c01000000171600" \
                      "14c7ed016bcf101160cd0b146a40133dc8eb6585fbffffffff87b40d41dac1f128fdf89e72ecda7fb5b286f5f586" \
                      "7164e3dd7a581ab53324270000000000ffffffff9b81806b10bf25e255164f026b9fff2ea9af0fe7b3a2ecdce974" \
                      "09b8f85d0e5a0000000000ffffffffdb7234b54fa8bb299248010e0062e141a42579f4e7642d258d1a889a0210c7" \
                      "490000000000ffffffff015838a21e00000000160014854d979a2739104f3476e15164aa6fcef0e950db02473044" \
                      "02206cc1163d3dfe22155c6479470391dcba86e3ccec7d783bb87bbce7cc47660ab102207c02779f21c835085e39" \
                      "24ff85541efa980cb350e1a3513b880b250fac491dd6012102b99f659fadd83993c4858d181922d5c9a8db74edb3" \
                      "6f2a7bbd84300a234d571b0247304402203eb46a7f5c84909291f81d5852421c62a6c19424343d4330f94a50eaf3" \
                      "01c31902200b9e406cd21613ba8022e3b2fe16bf00acf90b6faf8980ee56bd92c10bc2a8db012102d732900601aa" \
                      "2915532ff55ad8603f0c433eef9370ff9292b9696f2681588a720247304402201faecf60e30ebab7d3bbd05984ff" \
                      "f39c5c572fdf03567449a24152c24cdc1cab022078f9e0b338f26d2296c6882fb202ca73f7e229b441617951d879" \
                      "54c5881fb7eb01210376e28d103288b69b4e727ea384b9fa5a2833d68706b411f1cade00b681faea670248304502" \
                      "2100b0bc4018cf71e7d8f0af6b47e8c3bf531ed436abc946bc16fd466cd4f4d535ac02206f3706a9c423933855e6" \
                      "0ff876f4938aee30c0e81ed683c7e4c6286e97992ae0012103820065f5d287223d1c098db514c87e76d24089ca4c" \
                      "e5d4f208b4da97f2b5615800000000"

        expected_address = "tb1qs4xe0x388ygy7drku9gkf2n0emcwj5xmjxj7p8"
        decoded_expected_address = bech32decode(expected_address)
        deserialized = deserialize(transaction)
        assert decoded_expected_address == deserialized['outs'][0]['script']
        re_encoded_address = bech32encode(deserialized['outs'][0]['script'], prefix=BECH32_BITCOIN_TESTNET_PREFIX)
        assert re_encoded_address == expected_address

    def test_vectors(self):
        vectors = {
            "BC1QW508D6QEJXTDG4Y5R3ZARVARY0C5XW7KV8F3T4":
                "0014751e76e8199196d454941c45d1b3a323f1433bd6",

            "tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7":
                "00201863143c14c5166804bd19203356da136c985678cd4d27a1b8c6329604903262",

            "tb1qqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesrxh6hy":
                "0020000000c4a5cad46221b2a187905e5266362b99d5e91c6ce24d165dab93e86433"
        }
        for vector in vectors.items():
            assert bech32decode(vector[0]) == vector[1], (bech32decode(vector[0]), vector[1])

    def test_invalid_vectors(self):
        vectors = [
            'tc1qw508d6qejxtdg4y5r3zarvary0c5xw7kg3g4ty',
            'bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t5',
            'BC13W508D6QEJXTDG4Y5R3ZARVARY0C5XW7KN40WF2',
            'bc1rw5uspcuh',
            'bc10w508d6qejxtdg4y5r3zarvary0c5xw7kw508d6qejxtdg4y5r3zarvary0c5xw7kw5rljs90',
            'BC1QR508D6QEJXTDG4Y5R3ZARVARYV98GJ9P',
            'tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sL5k7',
            'bc1zw508d6qejxtdg4y5r3zarvaryvqyzf3du',
            'tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3pjxtptv',
            'bc1gmk9yu',
            'bc1pw508d6qejxtdg4y5r3zarvary0c5xw7kw508d6qejxtdg4y5r3zarvary0c5xw7k7grplx',
            'BC1SW50QA3JX3S',
            'bc1zw508d6qejxtdg4y5r3zarvaryvg6kdaj'
        ]
        with self.assertRaises(Exception):
            for vector in vectors:
                bech32decode(vector)

    def test_mktx(self):
        address = 'tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7'
        inputs = [
            {
                'output': 'ff'*32 + ':0',
                'value': 100,
                'segregated': True
            }
        ]
        outputs = [
            {
                'value': 100,
                'address': address
            }
        ]
        tx = mktx(inputs, outputs)
        deserialized = deserialize(tx)
        re_encoded_address = bech32encode(deserialized['outs'][0]['script'], prefix=BECH32_BITCOIN_TESTNET_PREFIX)
        assert re_encoded_address == address
        decoded = bech32decode(re_encoded_address)
        self.assertEqual(deserialized['outs'][0]['script'], decoded)

    def test_decode_signet_tx(self):
        tx = '020000000001010000000000000000000000000000000000000000000000000000000000000000ffffffff035a0101ff' \
             'ffffff020000000000000000160014a9c2890e5726fc19728a441a808728f130c51ff20000000000000000266a24aa21' \
             'a9ede2f61c3f71d1defd3fa999dfa36953755c690689799962b48bebd836974e8cf90120000000000000000000000000' \
             '000000000000000000000000000000000000000000000000'
        des = deserialize(tx)
        self.assertEqual(
            bech32encode(des['outs'][0]['script'], BECH32_CUSTOM_SIGNET_TESTNET_PREFIX),
            'cbt1q48pgjrjhym7pju52gsdgppeg7ycv28ljv3tr5h'
        )



from bitcoin.main import *
import hmac
import hashlib
# Electrum wallets


def electrum_stretch(seed):
    return slowsha(seed)

# Accepts seed or stretched seed, returns master public key


def electrum_mpk(seed):
    if len(seed) == 32:
        seed = electrum_stretch(seed)
    return privkey_to_pubkey(seed)[2:]

# Accepts (seed or stretched seed), index and secondary index
# (conventionally 0 for ordinary addresses, 1 for change) , returns privkey


def electrum_privkey(seed, n, for_change=0):
    if len(seed) == 32:
        seed = electrum_stretch(seed)
    mpk = electrum_mpk(seed)
    offset = dbl_sha256(from_int_representation_to_bytes(n)+b':'+from_int_representation_to_bytes(for_change)+b':'+binascii.unhexlify(mpk))
    return add_privkeys(seed, offset)

# Accepts (seed or stretched seed or master pubkey), index and secondary index
# (conventionally 0 for ordinary addresses, 1 for change) , returns pubkey


def electrum_pubkey(masterkey, n, for_change=0):
    if len(masterkey) == 32:
        mpk = electrum_mpk(electrum_stretch(masterkey))
    elif len(masterkey) == 64:
        mpk = electrum_mpk(masterkey)
    else:
        mpk = masterkey
    bin_mpk = encode_pubkey(mpk, 'bin_electrum')
    offset = bin_dbl_sha256(from_int_representation_to_bytes(n)+b':'+from_int_representation_to_bytes(for_change)+b':'+bin_mpk)
    return add_pubkeys('04'+mpk, privtopub(offset))

# seed/stretched seed/pubkey -> address (convenience method)


def electrum_address(masterkey, n, for_change=0, version=0):
    return pubkey_to_address(electrum_pubkey(masterkey, n, for_change), version)

# Given a master public key, a private key from that wallet and its index,
# cracks the secret exponent which can be used to generate all other private
# keys in the wallet


def crack_electrum_wallet(mpk, pk, n, for_change=0):
    bin_mpk = encode_pubkey(mpk, 'bin_electrum')
    offset = dbl_sha256(str(n)+':'+str(for_change)+':'+bin_mpk)
    return subtract_privkeys(pk, offset)


# Below code ASSUMES binary inputs and compressed pubkeys
MAINNET_PRIVATE = b'\x04\x88\xAD\xE4'
MAINNET_PUBLIC = b'\x04\x88\xB2\x1E'
TESTNET_PRIVATE = b'\x04\x35\x83\x94'
TESTNET_PUBLIC = b'\x04\x35\x87\xCF'
CUSTOM_SIGNET_TEST_PRIVATE = b'\x04\x35\x83\x82'
CUSTOM_SIGNET_TEST_PUBLIC = b'\x04\x35\x81\x84'
CUSTOM_SIGNET_PRIVATE = b'\x04\x88\xAD\x82'
CUSTOM_SIGNET_PUBLIC = b'\x04\x88\xB2\x84'

PRIVATE = [
    MAINNET_PRIVATE,
    TESTNET_PRIVATE,
    CUSTOM_SIGNET_TEST_PRIVATE,
    CUSTOM_SIGNET_PRIVATE
]
PUBLIC = [
    MAINNET_PUBLIC,
    TESTNET_PUBLIC,
    CUSTOM_SIGNET_TEST_PUBLIC,
    CUSTOM_SIGNET_PUBLIC
]


def get_corresponding_vbytes(vbytes: bytes):
    """
    return priv to pub or pub to priv vbytes
    """
    _ks = {
        MAINNET_PUBLIC: MAINNET_PRIVATE,
        MAINNET_PRIVATE: MAINNET_PUBLIC,
        TESTNET_PUBLIC: TESTNET_PRIVATE,
        TESTNET_PRIVATE: TESTNET_PUBLIC,
        CUSTOM_SIGNET_TEST_PUBLIC: CUSTOM_SIGNET_TEST_PRIVATE,
        CUSTOM_SIGNET_TEST_PRIVATE: CUSTOM_SIGNET_TEST_PUBLIC,
        CUSTOM_SIGNET_PUBLIC: CUSTOM_SIGNET_PRIVATE,
        CUSTOM_SIGNET_PRIVATE: CUSTOM_SIGNET_PUBLIC
    }
    return _ks[vbytes]


# BIP32 child key derivation


def raw_bip32_ckd(rawtuple, i):
    vbytes, depth, fingerprint, oldi, chaincode, key = rawtuple
    i = int(i)

    if vbytes in PRIVATE:
        priv = key
        pub = privtopub(key)
    elif vbytes in PUBLIC:
        priv = None
        pub = key
    else:
        raise ValueError('Unknown prefix')

    if i >= 2**31:
        if vbytes in PUBLIC:
            raise Exception("Can't do private derivation on public key!")

        I = hmac.new(chaincode, b'\x00'+priv[:32]+encode(i, 256, 4), hashlib.sha512).digest()
    else:
        I = hmac.new(chaincode, pub+encode(i, 256, 4), hashlib.sha512).digest()

    if vbytes in PRIVATE:
        newkey = add_privkeys(I[:32]+B'\x01', priv)
        fingerprint = bin_hash160(privtopub(key))[:4]
    else:
        newkey = add_pubkeys(compress(privtopub(I[:32])), key)
        fingerprint = bin_hash160(key)[:4]

    return vbytes, depth + 1, fingerprint, i, I[32:], newkey


def bip32_serialize(rawtuple):
    vbytes, depth, fingerprint, i, chaincode, key = rawtuple
    i = encode(i, 256, 4)
    chaincode = encode(hash_to_int(chaincode), 256, 32)
    keydata = b'\x00'+key[:-1] if vbytes in PRIVATE else key
    bindata = vbytes + from_int_to_byte(depth % 256) + fingerprint + i + chaincode + keydata
    return changebase(bindata+bin_dbl_sha256(bindata)[:4], 256, 58)


def bip32_deserialize(data):
    dbin = changebase(data, 58, 256)
    if bin_dbl_sha256(dbin[:-4])[:4] != dbin[-4:]:
        raise Exception("Invalid checksum")
    vbytes = dbin[0:4]
    depth = from_byte_to_int(dbin[4])
    fingerprint = dbin[5:9]
    i = decode(dbin[9:13], 256)
    chaincode = dbin[13:45]
    key = dbin[46:78]+b'\x01' if vbytes in PRIVATE else dbin[45:78]
    return vbytes, depth, fingerprint, i, chaincode, key


def raw_bip32_privtopub(rawtuple):
    vbytes, depth, fingerprint, i, chaincode, key = rawtuple
    newvbytes = get_corresponding_vbytes(vbytes)
    return newvbytes, depth, fingerprint, i, chaincode, privtopub(key)


def bip32_privtopub(data):
    return bip32_serialize(raw_bip32_privtopub(bip32_deserialize(data)))


def bip32_ckd(data, i):
    return bip32_serialize(raw_bip32_ckd(bip32_deserialize(data), i))


def bip32_master_key(seed, vbytes=MAINNET_PRIVATE):
    I = hmac.new(from_string_to_bytes("Bitcoin seed"), seed, hashlib.sha512).digest()
    return bip32_serialize((vbytes, 0, b'\x00'*4, 0, I[32:], I[:32]+b'\x01'))


def bip32_bin_extract_key(data):
    return bip32_deserialize(data)[-1]


def bip32_extract_key(data):
    return safe_hexlify(bip32_deserialize(data)[-1])

# Exploits the same vulnerability as above in Electrum wallets
# Takes a BIP32 pubkey and one of the child privkeys of its corresponding
# privkey and returns the BIP32 privkey associated with that pubkey


def raw_crack_bip32_privkey(parent_pub, priv):
    vbytes, depth, fingerprint, i, chaincode, key = priv
    pvbytes, pdepth, pfingerprint, pi, pchaincode, pkey = parent_pub
    i = int(i)

    if i >= 2**31:
        raise Exception("Can't crack private derivation!")

    I = hmac.new(pchaincode, pkey+encode(i, 256, 4), hashlib.sha512).digest()

    pprivkey = subtract_privkeys(key, I[:32]+b'\x01')

    newvbytes = MAINNET_PRIVATE if vbytes == MAINNET_PUBLIC else TESTNET_PRIVATE
    return newvbytes, pdepth, pfingerprint, pi, pchaincode, pprivkey


def crack_bip32_privkey(parent_pub, priv):
    dsppub = bip32_deserialize(parent_pub)
    dspriv = bip32_deserialize(priv)
    return bip32_serialize(raw_crack_bip32_privkey(dsppub, dspriv))


def coinvault_pub_to_bip32(*args):
    if len(args) == 1:
        args = args[0].split(' ')
    vals = map(int, args[34:])
    I1 = ''.join(map(chr, vals[:33]))
    I2 = ''.join(map(chr, vals[35:67]))
    return bip32_serialize((MAINNET_PUBLIC, 0, b'\x00'*4, 0, I2, I1))


def coinvault_priv_to_bip32(*args):
    if len(args) == 1:
        args = args[0].split(' ')
    vals = map(int, args[34:])
    I2 = ''.join(map(chr, vals[35:67]))
    I3 = ''.join(map(chr, vals[72:104]))
    return bip32_serialize((MAINNET_PRIVATE, 0, b'\x00'*4, 0, I2, I3+b'\x01'))


def bip32_descend(*args):
    if len(args) == 2 and isinstance(args[1], list):
        key, path = args
    else:
        key, path = args[0], map(int, args[1:])
    for p in path:
        key = bip32_ckd(key, p)
    return bip32_extract_key(key)



import unittest
from bitcoin import *


class TestTransaction(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Attempting transaction creation")

    # FIXME: I don't know how to write this as a unit test.
    # What should be asserted?
    def test_all(self):
        start = time.time()
        privs = [sha256(str(random.randrange(2**256))) for x in range(4)]
        pubs = [privtopub(priv) for priv in privs]
        addresses = [pubtoaddr(pub) for pub in pubs]
        mscript = mk_multisig_script(pubs[1:], 2, 3)
        msigaddr = p2sh_scriptaddr(mscript)
        tx = mktx(['01'*32+':1', '23'*32+':2'], [msigaddr+':20202', addresses[0]+':40404'], locktime=2222222222)
        tx1 = sign(tx, 1, privs[0])

        self.assertEqual(deserialize(tx)['locktime'], 2222222222, "Locktime incorrect")

        sig1 = multisign(tx, 0, mscript, privs[1])
        self.assertTrue(verify_tx_input(tx1, 0, mscript, sig1, pubs[1]), "Verification Error")

        sig3 = multisign(tx, 0, mscript, privs[3])
        self.assertTrue(verify_tx_input(tx1, 0, mscript, sig3, pubs[3]), "Verification Error")

        tx2 = apply_multisignatures(tx1, 0, mscript, [sig1, sig3])
        print('end %s' % (time.time() - start))

        from bitcoin import pubkey_to_address, script_to_address

    # https://github.com/vbuterin/pybitcointools/issues/71
    def test_multisig(self):
        start = time.time()
        script = mk_multisig_script(["0254236f7d1124fc07600ad3eec5ac47393bf963fbf0608bcce255e685580d16d9",
                                     "03560cad89031c412ad8619398bd43b3d673cb5bdcdac1afc46449382c6a8e0b2b"],
                                     2)

        self.assertEqual(p2sh_scriptaddr(script), "33byJBaS5N45RHFcatTSt9ZjiGb6nK4iV3")

        self.assertEqual(p2sh_scriptaddr(script, 0x05), "33byJBaS5N45RHFcatTSt9ZjiGb6nK4iV3")
        self.assertEqual(p2sh_scriptaddr(script, 5), "33byJBaS5N45RHFcatTSt9ZjiGb6nK4iV3")

        self.assertEqual(p2sh_scriptaddr(script, 0xc4), "2MuABMvWTgpZRd4tAG25KW6YzvcoGVZDZYP")
        self.assertEqual(p2sh_scriptaddr(script, 196), "2MuABMvWTgpZRd4tAG25KW6YzvcoGVZDZYP")
        print('end %s' % (time.time() - start))

    def test_preparetx(self):
        start = time.time()
        hextx = preparetx('12c6DSiU4Rq3P4ZxziKxzrL5LmMBrzjrJX', '1HLoD9E4SDFFPDiYfNYnkBLQ85Y51J3Zb1', 13)
        tx = deserialize(hextx)
        self.assertEqual(tx['locktime'], 0, "Locktime incorrect")
        self.assertEqual(tx['outs'][0]['value'], 13, "Value incorrect")

        hextx = preparetx('12c6DSiU4Rq3P4ZxziKxzrL5LmMBrzjrJX', '1HLoD9E4SDFFPDiYfNYnkBLQ85Y51J3Zb1', 13, locktime=2222222222)
        tx = deserialize(hextx)
        self.assertEqual(tx['locktime'], 2222222222, "Locktime incorrect")
        self.assertEqual(tx['outs'][0]['value'], 13, "Value incorrect")
        print('end %s' % (time.time() - start))


import hashlib
import os.path
import binascii
import random
from bisect import bisect_left

wordlist_english=list(open(os.path.join(os.path.dirname(os.path.realpath(__file__)),'english.txt'),'r'))

def eint_to_bytes(entint,entbits):
	a=hex(entint)[2:].rstrip('L').zfill(32)
	return binascii.unhexlify(a)

def mnemonic_int_to_words(mint,mint_num_words,wordlist=wordlist_english):
	backwords=[wordlist[(mint >> (11*x)) & 0x7FF].strip() for x in range(mint_num_words)]	
	return backwords[::-1]
	
def entropy_cs(entbytes):
	entropy_size=8*len(entbytes)
	checksum_size=entropy_size//32
	hd=hashlib.sha256(entbytes).hexdigest()
	csint=int(hd,16) >> (256-checksum_size)
	return csint,checksum_size
	
def entropy_to_words(entbytes,wordlist=wordlist_english):
	if(len(entbytes) < 4 or len(entbytes) % 4 != 0):
		raise ValueError("The size of the entropy must be a multiple of 4 bytes (multiple of 32 bits)")
	entropy_size=8*len(entbytes)
	csint,checksum_size = entropy_cs(entbytes)
	entint=int(binascii.hexlify(entbytes),16)
	mint=(entint << checksum_size) | csint
	mint_num_words=(entropy_size+checksum_size)//11
	
	return mnemonic_int_to_words(mint,mint_num_words,wordlist)

def words_bisect(word,wordlist=wordlist_english):
	lo=bisect_left(wordlist,word)
	hi=len(wordlist)-bisect_left(wordlist[:lo:-1],word)
	
	return lo,hi

def words_split(wordstr,wordlist=wordlist_english):
	def popword(wordstr,wordlist):
		for fwl in range(1,9):
			w=wordstr[:fwl].strip()
			lo,hi=words_bisect(w,wordlist)
			if(hi-lo == 1):
				return w,wordstr[fwl:].lstrip()
			wordlist=wordlist[lo:hi]
		raise Exception("Wordstr %s not found in list" %(w))

	words=[]
	tail=wordstr
	while(len(tail)):
		head,tail=popword(tail,wordlist)
		words.append(head)
	return words

def words_to_mnemonic_int(words,wordlist=wordlist_english):
	if(instance(words,str)):
		words=words_split(words,wordlist)
	return sum([wordlist.index(w) << (11*x) for x,w in enumerate(words[::-1])])

def words_verify(words,wordlist=wordlist_english):
	if(isinstance(words,str)):
		words=words_split(words,wordlist)
	
	mint = words_to_mnemonic_int(words,wordlist)
	mint_bits=len(words)*11
	cs_bits=mint_bits//32
	entropy_bits=mint_bits-cs_bits
	eint=mint >> cs_bits
	csint=mint & ((1 << cs_bits)-1)
	ebytes=_eint_to_bytes(eint,entropy_bits)
	return csint == entropy_cs(ebytes)

def mnemonic_to_seed(mnemonic_phrase,passphrase=u''):
	try:
		from hashlib import pbkdf2_hmac
		def pbkdf2_hmac_sha256(password,salt,iters=2048):
			return pbkdf2_hmac(hash_name='sha512',password=password,salt=salt,iterations=iters)
	except:
		try:
			from Crypto.Protocol.KDF import PBKDF2
			from Crypto.Hash import SHA512,HMAC
		
			def pbkdf2_hmac_sha256(password,salt,iters=2048):
				return PBKDF2(password=password,salt=salt,dkLen=64,count=iters,prf=lambda p,s: HMAC.new(p,s,SHA512).digest())
		except:
			try:
			
				from pbkdf2 import PBKDF2
				import hmac
				def pbkdf2_hmac_sha256(password,salt,iters=2048):
					return PBKDF2(password,salt, iterations=iters, macmodule=hmac, digestmodule=hashlib.sha512).read(64)
			except:
				raise RuntimeError("No implementation of pbkdf2 was found!")

	return pbkdf2_hmac_sha256(password=mnemonic_phrase,salt='mnemonic'+passphrase)

def words_mine(prefix,entbits,satisfunction,wordlist=wordlist_english,randombits=random.getrandbits):
	prefix_bits=len(prefix)*11
	mine_bits=entbits-prefix_bits
	pint=words_to_mnemonic_int(prefix,wordlist)
	pint<<=mine_bits
	dint=randombits(mine_bits)
	count=0
	while(not satisfunction(entropy_to_words(eint_to_bytes(pint+dint,entbits)))):
		dint=randombits(mine_bits)
		if((count & 0xFFFF) == 0):
			print("Searched %f percent of the space" % (float(count)/float(1 << mine_bits)))

	return entropy_to_words(eint_to_bytes(pint+dint,entbits))

if __name__=="__main__":
	import json
	testvectors=json.load(open('vectors.json','r'))
	passed=True
	for v in testvectors['english']:
		ebytes=binascii.unhexlify(v[0])
		w=' '.join(entropy_to_words(ebytes))
		seed=mnemonic_to_seed(w,passphrase='TREZOR')
		passed = passed and w==v[1]
		passed = passed and binascii.hexlify(seed)==v[2]
	print("Tests %s." % ("Passed" if passed else "Failed"))
		




#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name='conio-bitcoin',
      version='1.0.22',
      description='Python Bitcoin Tools',
      author='Vitalik Buterin',
      author_email='vbuterin@gmail.com',
      url='http://github.com/vbuterin/pybitcointools',
      packages=['bitcoin'],
      scripts=['pybtctool'],
      include_package_data=True,
      data_files=[("", ["LICENSE"]), ("bitcoin", ["bitcoin/english.txt"])],
      )



import unittest

from bitcoin import *


class TestP2TRTransaction(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Attempting P2TR transaction creation")

    def test2(self):
        pub = '512079be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798'
        addr = 'bc1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vqzk5jj0'
        self.assertEqual(bech32decode(addr), pub)
        self.assertEqual(pubkey_to_bech32_address(pub), addr)
        rtaddr = 'bcrt1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vqc8gma6'
        self.assertEqual(pubkey_to_bech32_address(pub, prefix=BECH32_BITCOIN_REGTEST_PREFIX), rtaddr)
        self.assertEqual(bech32decode(rtaddr), pub)


    def test(self):
        bech32_taproot_addr = 'tb1pqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesf3hn0c'
        priv = sha256(b'priv_key_text0rr')
        pub = compress(privtopub(priv))
        address = pubkey_to_bech32_address(pub, prefix='tb')
        print('depositaddr', address)
        transaction_to_sign = mktx(
            {
                'output': '341b1288a6e27673aeeff3a447218280550e71b2f0e4c86d301aa3cbab0b67f4:0',
                'segregated': True
            },
            [
                {'address': bech32_taproot_addr, 'value': int(0.001 * 10**8) - 5000}
            ]
        )
        tx = bech32_sign(transaction_to_sign, 0, priv, int(0.001 * 10**8))
        print(tx)
        self.assertEqual(segwit_txhash(tx)['txid'], '79a71c0e08b63b2f786b7649c765570e5aff93740eb295b5db4197390ceb4077')
